/*
   alloc.c - Memory allocation wrappers for libpdtp
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote 
 products derived from this software without specific prior written 
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT 
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* Prototype for the internal allocator */
static void *pdtp__internal_alloc(int size);
static void pdtp__internal_oom_handler(const char *allocator_name);

#ifdef WIN32
static void pdtp__internal_free(void *ptr);
#endif

/* Symbols for the internal allocator/deallocator/oom handler */
void *(*pdtp__alloc)(int size) = pdtp__internal_alloc;
#ifdef WIN32
void (*pdtp__free)(void *ptr) = pdtp__internal_free;
#else
void (*pdtp__free)(void *ptr) = free;
#endif /* WIN32 */
void (*pdtp__oom_handler)(const char *allocator_name) = pdtp__internal_oom_handler;

static void pdtp__internal_oom_handler(const char *allocator_name)
{
	perror(allocator_name);
	abort();
}

static void *pdtp__internal_alloc(int size)
{
	void *ret;

#ifdef WIN32
	if(!(ret = HeapAlloc(GetProcessHeap(), 0, size)))
		pdtp__oom_handler("HeapAlloc");
#else
	if(!(ret = malloc(size))) 
		pdtp__oom_handler("malloc");
#endif /* WIN32 */
	
	return ret;
}

#ifdef WIN32
static void pdtp__internal_free(void *ptr)
{
	HeapFree(GetProcessHeap(), 0, ptr);
}
#endif

void pdtp_alloc_set(void *(*function)(int))
{
	pdtp__alloc = function;
}

void pdtp_free_set(void (*function)(void *))
{
	pdtp__free = function;
}

void pdtp_oom_handler_set(void (*function)(const char *))
{
	pdtp__oom_handler = function;
}

char *pdtp__strdup(const char *str)
{
	char *ret = (char *)pdtp__alloc((unsigned)strlen(str) + 1);

	strcpy(ret, str);
	
	return ret;
}
