/*
   connection.c - Functions for establishing a PDTP connection
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote
 products derived from this software without specific prior written
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "sockets.h"
#include "pdtp.h"

#include "alloc.h"
#include "connection.h"
#include "collection.h"
#include "error.h"
#include "notifier.h"
#include "socketops.h"
#include "threadops.h"

/* Protocol magic numbers */
#define PDTP_MAGIC	0x50445450
#define PDRS_MAGIC	0x50445253

pdtp_connection_t pdtp_connect(const char *hostname, int port)
{
	int sock;
	pdtp_addr_t addr;

	uint16_t i16;
	uint32_t i32;
	pdtp_connection_t connection;

	if(!(addr = pdtp__resolve(hostname, port)))
		return 0;

	if((sock = pdtp__connect(addr)) < 0)
		goto err0;

	pdtp__writesock_int32(sock, PDTP_MAGIC);

	/* XXX Redirection should operate in a loop like so: */
	/* while(!pdtp__readsock_int32(sock, &i32) && i32 == PDRS_MAGIC) { */
	if(pdtp__readsock_int32(sock, &i32))
		goto err1;

	if(i32 == PDRS_MAGIC) {
		pdtp_set_last_error(ERROR_RDR_UNIMPL);

		/* XXX Implement redirection handler */
		goto err1;
	}

	if(i32 != PDTP_MAGIC) 
		goto err1;

	/* Read the minimum supported protocol version and check if it's 1 */
	if(pdtp__readsock_int16(sock, &i16) < 0 || i16 != 1)
		goto err1;

	/* Read the maximum supported protocol version */
	if(pdtp__readsock_int16(sock, &i16) < 0)
		goto err1;

	/* Select the protocol version, currently hardcoded to 1 */
	if(pdtp__writesock_int16(sock, 1) < 0)
		goto err1;

	connection = NEW(pdtp_connection);
	connection->sock = sock;
	connection->addr = addr;
	
	connection->socket_lock = pdtp__mutex_create();
	connection->txnmgr = pdtp__transaction_manager_create();

	pdtp__set_error_ptr(0);
	return connection;

err1:
	pdtp_set_last_error(ERROR_HANDSHAKE);
	pdtp__socket_close(sock);
err0:
	pdtp__addr_destroy(addr);
	
	return 0;
}

void pdtp_close(pdtp_connection_t connection)
{
	pdtp__addr_destroy(connection->addr);
	pdtp__socket_close(connection->sock);

	pdtp__mutex_destroy(connection->socket_lock);
	pdtp__transaction_manager_destroy(connection->txnmgr);

	pdtp__free(connection);
}
