/*
   directory.c - Functions for performing directory listings
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote
 products derived from this software without specific prior written
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "pdtp.h"

#include <string.h>

#include "alloc.h"
#include "connection.h"
#include "error.h"
#include "transaction.h"
#include "txid.h"

struct pdtp_directory {
	pdtp_transaction_in_t txn;
	int i, n;
};

pdtp_directory_t pdtp_get_directory_listing(pdtp_connection_t conn, const char *path, pdtp_read_mode_t mode)
{
	int index;
	char *errstr;
	pdtp_transaction_out_t out;
	pdtp_transaction_in_t in;
	pdtp_notifier_t notifier;
	pdtp_directory_t dir;

	/* Send a request for a directory listing */
	out = pdtp__transaction_create(TXN_DIRLIST, 1);
	notifier = pdtp__transaction_set_serial(out, conn);
	pdtp__transaction_add_string(out, OBJ_PATH, path);
	pdtp__transaction_write(conn, out);
	pdtp__transaction_out_destroy(out);

	/* Retrieve the transaction by the notifier interface */
	if(!(in = pdtp__transaction_wait(conn, notifier, mode))) {
		/* XXX The connection was lost, handle this */
		return 0;
	}

	if(!pdtp__transaction_id(in)) {
		if((index = pdtp__transaction_object_index(in, OBJ_ERRMSG)) < 0 || pdtp__transaction_object_string(in, index, &errstr) < 0)
			pdtp_set_last_error(ERROR_DIRLIST);
		else 
			pdtp__set_error_ptr(errstr);

		pdtp__transaction_in_destroy(in);
		return 0;
	}

	dir = NEW(pdtp_directory);
	dir->txn = in;
	dir->i = 0;
	dir->n = pdtp__transaction_object_count(in);

	printf("Returning directory listing with %d items\n", dir->n);

	return dir;
}

void pdtp_directory_close(pdtp_directory_t dir)
{
	pdtp__transaction_in_destroy(dir->txn);
	pdtp__free(dir);
}

void pdtp_directory_rewind(pdtp_directory_t dir)
{
	dir->i = 0;
}

int pdtp_get_direntry(pdtp_directory_t dir, pdtp_direntry_t *entry)
{
	uint16_t id;
	char *str;

	if(dir->i == dir->n)
		return 1;

	if(pdtp__transaction_object_id(dir->txn, dir->i, &id) < 0) 
		return -1;
	
	switch(id) { 
		case OBJ_FILE: 
			if(pdtp__transaction_object_unpack(dir->txn, dir->i, "248s", &entry->ts_year, &entry->ts_ysec, &entry->length, &str) != 4)
				return -1;
			
			entry->type = PDTP_FILE;
			break;
		case OBJ_SUBDIR:
			if(pdtp__transaction_object_unpack(dir->txn, dir->i, "24s", &entry->ts_year, &entry->ts_ysec, &str) != 3) 
				return -1;

			entry->type = PDTP_DIR;
			entry->length = 0;
			break;
	}

	strncpy(entry->name, str, PDTP_PATH_MAX);
	pdtp__free(str);

	dir->i++;

	return 0;
}
