/*
   download.c - High level download control interface
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote 
 products derived from this software without specific prior written 
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT 
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "pdtp.h"
#include "sockets.h"

#include <stdio.h>
#include <string.h>

#include "alloc.h"
#include "connection.h"
#include "debug.h"
#include "errmsg.h"
#include "error.h"
#include "fileops.h"
#include "listener.h"
#include "multiplexer.h"
#include "piecelist.h"
#include "transaction.h"
#include "txid.h"

/* Amount of data to be read/written at a time when transferring a piece */
#define TRANSFER_CHUNK_SIZE	4096

/* Internal definition of download structure */
struct pdtp_download {
        /* Handle of file */
        uint32_t handle;
		
	/* Length of file */
	uint64_t length;

	/* Piece length */
	uint32_t piece_length;

	/* List of hashes */
	uint8_t *hashes;
	
	/* Connection associated with download */
	pdtp_connection_t conn;

	/* Multiplexer associated with download */
	pdtp_multiplexer_t mplx;

	/* File being downloaded */
	pdtp_file_t file;

	/* Callback table associated with download */
	pdtp_callback_table_t *callback;
};

/* This structure stores piece transfer initialization parameters */
typedef struct piece_transfer_parameters {
	pdtp_download_t dl;
	uint32_t peer_address;
	uint16_t opcode;
	uint32_t piece_index;
	unsigned bytes_transferred, bytes_total;
	uint8_t started;
} *piece_transfer_parameters_t;

/* This structure stores incoming transfer parameters */
typedef struct incoming_connection_parameters {
	int fd;
	piece_transfer_parameters_t params;
} *incoming_connection_parameters_t;

/* Event handler prototypes */
static int incoming_transaction_handler(pdtp_multiplexer_t mplx, pdtp_transaction_listener_t listener, pdtp_transaction_in_t txn, void *arg);
int incoming_connection_handler(pdtp_multiplexer_t mplx, int fd, uint16_t opcode, uint32_t addr, uint16_t port, uint32_t handle, uint32_t piece_index, void *arg);

/* The high level interface is defined first.  Event handlers and actual
   mechanics of the download engine follow below as a series of static
   functions which are internal to this module. */
pdtp_download_t pdtp_download_create(pdtp_connection_t conn, const char *path)
{
	unsigned i, piece_count;
	uint32_t length;
	void *data;
	char *errstr;
	
	pdtp_transaction_out_t out;
	pdtp_transaction_in_t in;
	pdtp_notifier_t notifier;
	pdtp_download_t dl;

	/* Send a request for file information */
	out = pdtp__transaction_create(TXN_FILEINFO, 1);
	notifier = pdtp__transaction_set_serial(out, conn);
	pdtp__transaction_add_string(out, OBJ_PATH, path);
	pdtp__transaction_write(conn, out);
	pdtp__transaction_out_destroy(out);

	/* Read the server's reply */
	if(!(in = pdtp__transaction_wait(conn, notifier, PDTP_STORE)))
		return 0;

	if(!pdtp__transaction_id(in)) {
		if((i = pdtp__transaction_object_index(in, OBJ_ERRMSG)) < 0 || pdtp__transaction_object_string(in, i, &errstr) < 0)
			pdtp_set_last_error(ERROR_DOWNLOAD_INIT);
		else
			pdtp__set_error_ptr(errstr);

		pdtp__transaction_in_destroy(in);
		return 0;
	}
		
	dl = NEW(pdtp_download);

	if((i = pdtp__transaction_object_index(in, OBJ_HANDLE)) < 0)
		goto err;

	if(pdtp__transaction_object_uint32(in, i, &dl->handle) < 0)
		goto err;

#ifndef NDEBUG
	debug("dl->handle is: %d", dl->handle);
#endif
	
	if((i = pdtp__transaction_object_index(in, OBJ_FILELEN)) < 0)
		goto err;
	
	if(pdtp__transaction_object_uint64(in, i, &dl->length) < 0)
		goto err;

#ifndef NDEBUG
	debug("dl->length is: %d", (int)dl->length);
#endif
	
	if((i = pdtp__transaction_object_index(in, OBJ_PIECELEN)) < 0)
		goto err;

	if(pdtp__transaction_object_uint32(in, i, &dl->piece_length) < 0)
		goto err;

	piece_count = (unsigned)(dl->length / dl->piece_length);
	if(dl->length % dl->piece_length)
		piece_count++;

	if((i = pdtp__transaction_object_index(in, OBJ_MD5HASH)) < 0)
		goto err;

	/* Check if the hashes block is the expected length */
	if(pdtp__transaction_object_length(in, i, &length) < 0)
		goto err;
	
	if(piece_count * 16 != length)
		goto err;

	if(pdtp__transaction_object_data(in, i, &data) < 0)
		goto err;
	
	dl->hashes = (uint8_t *)pdtp__alloc(length);
	
	memcpy(dl->hashes, data, length);

	pdtp__transaction_in_destroy(in);

	dl->conn = conn;
	dl->mplx = 0;
	dl->file = 0;
	dl->callback = 0;

	return dl;
err:
	pdtp__transaction_in_destroy(in);
	pdtp__free(dl);

	return 0;
}

void pdtp_download_destroy(pdtp_download_t dl)
{
	pdtp__free(dl->hashes);
	pdtp__free(dl);
}

void pdtp_download_set_option(pdtp_download_t dl, pdtp_download_option_t option, int value)
{
}

void pdtp_download_run(pdtp_download_t dl, pdtp_file_t file, pdtp_callback_table_t *table, void *ptr)
{
	uint32_t value;
	pdtp_piecelist_t list;
	pdtp_transaction_in_t in;
	pdtp_transaction_out_t out;
	pdtp_notifier_t notifier;

	/* pdtp_piecelist_create() handles not only MD5 checksumming
	   but also truncating the file to the specified length */
	if(!(list = pdtp__piecelist_create(file, (pdtp_filelen_t)dl->length, dl->piece_length, dl->hashes))) {
		/* XXX Some sort of error handling here */

		return;
	}

	out = pdtp__transaction_create(TXN_DOWNLOAD, 2);
	notifier = pdtp__transaction_set_serial(out, dl->conn);
	pdtp__transaction_add_int32(out, OBJ_HANDLE, dl->handle);
	pdtp__piecelist_add_to_transaction(out, list);
	pdtp__transaction_write(dl->conn, out);
	pdtp__transaction_out_destroy(out);
	pdtp__piecelist_destroy(list);

	if(!(in = pdtp__transaction_wait(dl->conn, notifier, PDTP_STORE))) {
#ifndef NDEBUG
		debug("pdtp_download_run: pdtp__transaciton_wait returned NULL");
#endif
		return;
	}

	if(!pdtp__transaction_id(in)) {
#ifndef NDEBUG
		debug("pdtp_download_run: Server error beginning download");
#endif
		/* XXX Fix error handling */
		
		pdtp__transaction_in_destroy(in);
		return;
	}
	
	pdtp__transaction_in_destroy(in);

	/* Create the multiplexer for this download */
	dl->mplx = pdtp__multiplexer_create();

#ifndef NDEBUG
	debug("Scanning for handle: %d", dl->handle);
#endif

	/* Convert value to big endian */
	value = htonl(dl->handle);
	
	/* Create the notifier to receive download events */
	notifier = pdtp__create_object_notifier(dl->conn, OBJ_HANDLE, &value, 4);

	/* Create the transaction listener and add it to the multiplexer */
	pdtp__multiplexer_add_transaction_listener(dl->mplx, pdtp__transaction_listener_create(dl->conn, notifier, PDTP_STORE), incoming_transaction_handler, dl);

	/* Create the notifier for the listener socket */
	if((notifier = pdtp__listener_notifier_create(pdtp__addr_ipv4(dl->conn->addr), pdtp__addr_port(dl->conn->addr), dl->handle)))
		pdtp__multiplexer_add_ipv4_listener(dl->mplx, notifier, incoming_connection_handler, dl);

	dl->file = file;
        dl->callback = table;

	pdtp__multiplexer_run(dl->mplx);
}

void pdtp_download_stop(pdtp_download_t dl)
{
}

static void transfer_success_notify(piece_transfer_parameters_t params)
{
        pdtp_transaction_out_t txn;

        txn = pdtp__transaction_create(TXN_XFERSUCCESS, 3);
        pdtp__transaction_add_int32(txn, OBJ_HANDLE, params->dl->handle);
        pdtp__transaction_add_int32(txn, OBJ_PIECEID, params->piece_index);
        pdtp__transaction_add_int32(txn, OBJ_IPV4ADDR, params->peer_address);
        pdtp__transaction_write(params->dl->conn, txn);
        pdtp__transaction_out_destroy(txn);
}

static void transfer_failure_notify(piece_transfer_parameters_t params)
{
	pdtp_transaction_out_t txn;

	txn = pdtp__transaction_create(TXN_XFERFAIL, 3);
	pdtp__transaction_add_int32(txn, OBJ_HANDLE, params->dl->handle);
	pdtp__transaction_add_int32(txn, OBJ_PIECEID, params->piece_index);
	pdtp__transaction_add_int32(txn, OBJ_IPV4ADDR, params->peer_address);
	pdtp__transaction_write(params->dl->conn, txn);
	pdtp__transaction_out_destroy(txn);
}

static int peer_upload_piece(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg)
{
	int nbytes;
	uint8_t chunk[TRANSFER_CHUNK_SIZE];

	piece_transfer_parameters_t params = (piece_transfer_parameters_t)arg;
	
	if(bytes < 0) 
		goto err;
	
	/* Check and set flag for if the transfer has started */
	if(!params->started)
		params->started = 1;
	else 
		params->bytes_transferred += bytes;

	if(params->bytes_transferred == params->bytes_total) {
		transfer_success_notify(params);
		goto done;
	}

	/* XXX Input validation on the piece index should occur earlier on */
	if(pdtp__file_seek(params->dl->file, (uint64_t)params->dl->piece_length * params->piece_index + params->bytes_transferred) < 0)
		goto err;

	/* XXX Implement proper calculation of the number of bytes to read */
	if((nbytes = pdtp__file_read(params->dl->file, chunk, TRANSFER_CHUNK_SIZE)) < 0)
		goto err;

	pdtp__nonblock_write(mplx, sock, chunk, nbytes, peer_upload_piece, params);

	return 1;

err:
	transfer_failure_notify(params);
done:
	/* XXX Add support for persistent connections */
	pdtp__free(params);
	pdtp__socket_close(sock);

	return 1;
}

static int peer_download_piece(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg)
{ 
	piece_transfer_parameters_t params = (piece_transfer_parameters_t)arg;

        if(bytes < 0)
                goto err;

	/* XXX Add proper size calculation here */
        if(!params->started) {
                params->started = 1;
		goto read_next_chunk;
	}

	/* XXX Input validation on the piece index should occur earlier on */
        if(pdtp__file_seek(params->dl->file, (uint64_t)params->dl->piece_length * params->piece_index + params->bytes_transferred) < 0)
                goto err;

	if(pdtp__file_write(params->dl->file, buffer, bytes) < 0)
		goto err;

	params->bytes_transferred += bytes;
	
	if(params->bytes_transferred == params->bytes_total) { 
		transfer_success_notify(params);
		goto done; 
	}
	
read_next_chunk:
	/* XXX Add proper size calculation here */
	pdtp__nonblock_read(mplx, sock, TRANSFER_CHUNK_SIZE, peer_download_piece, params);

	goto done;
err:
	transfer_failure_notify(params);
done:
        pdtp__free(params);
        pdtp__socket_close(sock);

	return 1;
}

static int peer_send_request(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg)
{
	piece_transfer_parameters_t params = (piece_transfer_parameters_t)arg;
	int (*transfer_function)(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg) = 0;
	uint16_t v16;
	uint32_t v32;
	
	uint8_t request[18];

	if(bytes < 0)
		goto err;

	/* Check the magic number */
	if(memcmp(buffer, "PTPP", 4))
		goto err;

	/* Check if the minimum supported protocol version is 1 */
	memcpy(&v16, (uint8_t *)buffer + 4, 2);

	if(htons(v16) != 1)
		goto err;

	/* Select the protocol version.  Currently hardcoded to 1 */
	v16 = htons(1);
	memcpy(request, &v16, 2);

	/* Set operation code. */
	v16 = htons(params->opcode);
	memcpy(request + 2, &v16, 2);

	/* Set address of coordinating server */
	v32 = htonl(pdtp__addr_ipv4(params->dl->conn->addr));
	memcpy(request + 4, &v32, 4);

	/* Set port of coordinating server.  Initialize properly */
	v16 = htons(pdtp__addr_port(params->dl->conn->addr));
	memcpy(request + 8, &v16, 4);

	/* Set file handle ID. */
	v32 = htonl(params->dl->handle);
	memcpy(request + 10, &v32, 4);

	/* Set piece ID. */
	v32 = htonl(params->piece_index);
	memcpy(request + 14, &v32, 4);
	
	switch(params->opcode) {
		case TXN_SENDPIECE:
			transfer_function = peer_upload_piece;
			break;
		case TXN_RECVPIECE:
			transfer_function = peer_download_piece;
			break;
	}

	pdtp__nonblock_write(mplx, sock, request, 18, transfer_function, params);
	
	return 1;

err:
	transfer_failure_notify(params);
	pdtp__free(params);
	pdtp__socket_close(sock);
	return -1;
}

static int peer_read_header(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg)
{
	/* If a write error occured terminate the connection and notify server*/
	if(bytes < 0) {
		transfer_failure_notify((piece_transfer_parameters_t)arg);
		pdtp__free(arg);
		pdtp__socket_close(sock);
		return 1;
	}

	pdtp__nonblock_read(mplx, sock, 8, peer_send_request, arg);

	return 1;
}

static int connect_complete_handler(pdtp_multiplexer_t mplx, int sock, void *arg)
{
	pdtp__multiplexer_remove_socket(mplx, sock);

	/* Check to see whether the connection attempt was successful */
	if(pdtp__nonblock_connect_complete(sock) < 0) {
		transfer_failure_notify((piece_transfer_parameters_t)arg);
		pdtp__free(arg);
		pdtp__socket_close(sock);
		return 1;
	}

	pdtp__nonblock_write(mplx, sock, "PTPP", 4, peer_read_header, arg);

	return 1;
}

static int incoming_transaction_handler(pdtp_multiplexer_t mplx, pdtp_transaction_listener_t listener, pdtp_transaction_in_t txn, void *arg)
{
	pdtp_sock_t sock;
	unsigned i, id = pdtp__transaction_id(txn);
	uint32_t ip, piece_index;
	uint16_t port;
	pdtp_addr_t addr;
	piece_transfer_parameters_t params;
	pdtp_download_t dl = (pdtp_download_t)arg;

#ifndef NDEBUG
	debug("Handling transaction: %d", id);
#endif

	/* This needs to accomidate more transactions */
	if(id != TXN_SENDPIECE && id != TXN_RECVPIECE) {
#ifndef NDEBUG
		debug("Received bogus transaction through download filter");
#endif
		goto err;
	}
		
	if((i = pdtp__transaction_object_index(txn, OBJ_PIECEID)) < 0)
		goto err;

	if(pdtp__transaction_object_uint32(txn, i, &piece_index) < 0)
		goto err;

	if((i = pdtp__transaction_object_index(txn, OBJ_IPV4ADDR)) < 0)
		goto err;

	if(pdtp__transaction_object_uint32(txn, i, &ip) < 0)
		goto err;

	if((i = pdtp__transaction_object_index(txn, OBJ_PORT)) < 0)
		goto err;

	if(pdtp__transaction_object_uint16(txn, i, &port) < 0)
		goto err;

	addr = pdtp__ipv4_addr_create(ip, port);
	sock = pdtp__nonblock_connect(addr);
	pdtp__addr_destroy(addr);

	if(sock < 0)
		goto err;

	params = NEW(piece_transfer_parameters);
	params->started = 0;
	params->dl = dl;
	params->peer_address = ip;
	params->opcode = id;
	params->piece_index = piece_index;
	params->bytes_transferred = 0;

	/* XXX Initialize this value properly */
	params->bytes_total = dl->piece_length;

	pdtp__multiplexer_add_socket(mplx, sock, MPLX_WRITE, connect_complete_handler, params, 0);
	
	pdtp__transaction_in_destroy(txn);

	return 1;

err:
	pdtp__transaction_in_destroy(txn);
	return 0;
}

int transfer_authorization_handler(pdtp_multiplexer_t mplx, pdtp_transaction_listener_t listener, pdtp_transaction_in_t txn, void *arg)
{
	incoming_connection_parameters_t cxnp = (incoming_connection_parameters_t)arg;
	piece_transfer_parameters_t params = cxnp->params;
	int fd = cxnp->fd;
	int id = pdtp__transaction_id(txn);

	pdtp__free(cxnp);
	pdtp__transaction_in_destroy(txn);

	if(!id) {
		pdtp__socket_close(fd);
		pdtp__free(params);
		return 1;
	}
	
	switch(params->opcode) { 
		case TXN_SENDPIECE: 
			return peer_download_piece(mplx, fd, 0, 0, params);
		case TXN_RECVPIECE: 
			return peer_upload_piece(mplx, fd, 0, 0, params);
	}
	
	return 1;
}

int incoming_connection_handler(pdtp_multiplexer_t mplx, int fd, uint16_t opcode, uint32_t addr, uint16_t port, uint32_t handle, uint32_t piece_index, void *arg)
{
	uint16_t id;
	pdtp_transaction_out_t out;
	pdtp_notifier_t notifier;
	pdtp_transaction_listener_t listener;
	pdtp_download_t dl = (pdtp_download_t)arg;
	piece_transfer_parameters_t params;
	incoming_connection_parameters_t cxnp;

#ifndef NDEBUG
	debug("Incoming connection from %u.%u.%u.%u", ((uint8_t *)&addr)[0], ((uint8_t *)&addr)[1], ((uint8_t *)&addr)[2], ((uint8_t *)&addr)[3]);
#endif
	
	switch(opcode) {
		case TXN_SENDPIECE:
			id = TXN_AUTHDL;
			break;
		case TXN_RECVPIECE:
			id = TXN_AUTHUL;
			break;
		default:
#ifndef NDEBUG
			debug("Incoming connection sent bad opcode: %d", opcode);
#endif
			return 1;
	}
	
	out = pdtp__transaction_create(id, 3);
        notifier = pdtp__transaction_set_serial(out, dl->conn);
        pdtp__transaction_add_int32(out, OBJ_HANDLE, dl->handle);
	pdtp__transaction_add_int32(out, OBJ_PIECEID, piece_index);
	pdtp__transaction_add_int32(out, OBJ_IPV4ADDR, addr);
        pdtp__transaction_write(dl->conn, out);
        pdtp__transaction_out_destroy(out);

	listener = pdtp__transaction_listener_create(dl->conn, notifier, PDTP_STORE);

        params = NEW(piece_transfer_parameters);
        params->started = 0;
        params->dl = dl;
        params->peer_address = addr;
        params->opcode = opcode;
        params->piece_index = piece_index;
        params->bytes_transferred = 0;
        params->bytes_total = dl->piece_length; /* XXX Initialize properly */

	cxnp = NEW(incoming_connection_parameters);
	cxnp->fd = fd;
	cxnp->params = params;

	pdtp__multiplexer_add_transaction_listener(mplx, listener, transfer_authorization_handler, cxnp);
			
	return 1;
}
