/*
   error.c - Thread-specific error messages
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote
 products derived from this software without specific prior written
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "pdtp.h"

#include <stdio.h>
#include <stdarg.h>

#include "alloc.h"

/* Win32 implemenatation using Thread Local Storage */
#ifdef WIN32
static DWORD error_message_key;

/* Hopefully we have no other use for DllMain, otherwise this code will need 
   to be moved elsewhere... */
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{ 
	LPVOID ptr;

	switch (fdwReason) 
	{ 
		case DLL_PROCESS_ATTACH: 
			/* Allocate a TLS index */
			if ((error_message_key = TlsAlloc()) == 0xFFFFFFFF) 
				return FALSE; 

			/* Fall through to initialize the parent thread */
		case DLL_THREAD_ATTACH: 
			TlsSetValue(error_message_key, 0); 
			break; 
		case DLL_THREAD_DETACH:
			/* Free the error message if it's still allocated */
			if((ptr = TlsGetValue(error_message_key)))
				pdtp__free(ptr);

			break; 
		case DLL_PROCESS_DETACH: 
			/* Free the error message if it's still allocated */
			if((ptr = TlsGetValue(error_message_key))) 
				pdtp__free(ptr);

			/* Release the TLS index. */
			TlsFree(error_message_key); 

			break; 
	} 

	return TRUE; 

	UNREFERENCED_PARAMETER(hinstDLL); 
	UNREFERENCED_PARAMETER(lpvReserved); 
}

void pdtp__set_error_ptr(char *message)
{
	LPVOID ptr;

	if((ptr = TlsGetValue(error_message_key)))
		pdtp__free(ptr);

	TlsSetValue(error_message_key, message);
}

char *pdtp_get_last_error(void)
{
	return (char *)TlsGetValue(error_message_key);
}
#endif /* WIN32 */

/* POSIX threads implementation using thread specific data */
#ifdef HAVE_PTHREADS
#include <pthread.h>

static pthread_key_t error_message_key;
static pthread_once_t error_message_once = PTHREAD_ONCE_INIT;

static void error_message_destructor(void *ptr)
{
	if(ptr) 
		pdtp__free(ptr);
}

static void error_message_init()
{
	pthread_key_create(&error_message_key, error_message_destructor);
	pthread_setspecific(error_message_key, 0);
}

void pdtp__set_error_ptr(char *message)
{
	void *ptr;

	pthread_once(&error_message_once, error_message_init);

	if((ptr = pthread_getspecific(error_message_key)))
		pdtp__free(ptr);

	pthread_setspecific(error_message_key, message);
}

char *pdtp_get_last_error(void)
{
	char *ret;

	pthread_once(&error_message_once, error_message_init);

	ret = (char *)pthread_getspecific(error_message_key);
	pthread_setspecific(error_message_key, 0);

	return ret;
}
#endif /* HAVE_PTHREADS */

/* Non-thread safe implementation */
#if !defined(WIN32) && !defined(HAVE_PTHREADS)
#warning Building with non-thread safe version of pdtp_get_last_error()
static char *error_message = 0;

void pdtp__set_error_ptr(char *message)
{
	if(error_message)
		pdtp__free(error_message);

	error_message = message;
}

char *pdtp_get_last_error(void)
{
	char *ret = error_message;
	error_message = 0;

	return ret;
}
#endif /* !defined(WIN32) && !defined(HAVE_PTHREADS) */

void pdtp_set_last_error(const char *message)
{
	if(message) 
		pdtp__set_error_ptr(pdtp__strdup(message));
	else
		pdtp__set_error_ptr(0);
}

/* This is the default buffer size to allocate for systems lacking vasprintf. */
#define INITIAL_BUFFER 128

/* On Windows, if a _vsnprintf() call fails we keep doubling the buffer size
   until we hit this sanity limit, to prevent an infinite allocation loop */
#define MAX_BUFFER 4096

void pdtp__error_printf(const char *format, ...)
{
	char *msg;
	va_list ap;

#ifdef WIN32
	int size = INITIAL_BUFFER;

	msg = (char *)pdtp__alloc(size);
	va_start(ap, format);

	while(_vsnprintf(msg, size, format, ap) < 0) {
		va_end(ap);
		pdtp__free(msg);
		
		if((size *= 2) > MAX_BUFFER)
			return;

		msg = (char *)pdtp__alloc(size);
		va_start(ap, format);
	}

	pdtp__set_error_ptr(msg);
#endif /* WIN32 */
	
#ifdef HAVE_VASPRINTF
	va_start(ap, format);
	if(vasprintf(&msg, format, ap) < 0) 
		pdtp__oom_handler("vasprintf");
	va_end(ap);

	pdtp__set_error_ptr(msg);
#endif /* HAVE_VASPRINTF */

#if !defined(WIN32) && !defined(HAVE_VASPRINTF)
	int l;

	msg = (char *)pdtp__alloc(INITIAL_BUFFER);

	va_start(ap, format);
	l = vsnprintf(msg, INITIAL_BUFFER, format, ap);
	va_end(ap);

	if(l < INITIAL_BUFFER) {
		pdtp__set_error_ptr(msg);
		return;
	}

	l++;
	pdtp__free(msg);
	msg = (char *)pdtp__alloc(l);

	va_start(ap, format);
        vsnprintf(msg, l, format, ap);
	va_end(ap);

	pdtp__set_error_ptr(msg);
#endif /* !defined(WIN32) && !defined(HAVE_VASPRINTF) */
}
