/*
   fileops.c - Portable wrappers around file operations
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote 
 products derived from this software without specific prior written 
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT 
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "pdtp.h"

#ifdef WIN32
#ifndef INVALID_SET_FILE_POINTER
#define INVALID_SET_FILE_POINTER ((DWORD)-1)
#endif /* INVALID_SET_FILE_POINTER */
#else
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#endif /* WIN32 */

#include "fileops.h"

int pdtp_fopen(char *filename, pdtp_file_t *file)
{
#ifdef WIN32
	HANDLE fh;
	
	if((fh = CreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, 0)) == INVALID_HANDLE_VALUE)
		return -1;

	*file = fh;
#else
	int fd;

	if((fd = open(filename, O_RDWR | O_CREAT, 0644)) < 0)
		return -1;

	*file = fd;
#endif /* WIN32 */

	return 0;
}

void pdtp_fclose(pdtp_file_t file)
{
#ifdef WIN32
	CloseHandle(file);
#else
	close(file);
#endif /* WIN32 */
}

int pdtp__file_length(pdtp_file_t file, pdtp_filelen_t *length)
{
#ifdef WIN32
	ULARGE_INTEGER fsize;

	if((fsize.LowPart = GetFileSize(file, &fsize.HighPart)) == INVALID_FILE_SIZE)
		return -1;

	*length = fsize.QuadPart;
#else
	struct stat st;

	if(fstat(file, &st) < 0)
		return -1;

	*length = st.st_size;
#endif /* WIN32 */

	return 0;
}

int pdtp__file_truncate(pdtp_file_t file, pdtp_filelen_t length)
{
#ifdef WIN32
	LARGE_INTEGER fsize;

	fsize.QuadPart = length;

	/* Set the pointer to the length we wish to truncate to */
	if(SetFilePointer(file, fsize.LowPart, &fsize.HighPart, FILE_BEGIN) == INVALID_SET_FILE_POINTER && GetLastError() != NO_ERROR)
		return -1;

	/* Truncate the file */
	if(!SetEndOfFile(file))
		return -1;
#else
	if(ftruncate(file, length) < 0) 
		return -1;
#endif /* WIN32 */

	return 0;
}

int pdtp__file_seek(pdtp_file_t file, pdtp_filelen_t offset)
{
#ifdef WIN32
	LARGE_INTEGER fsize;

	fsize.QuadPart = offset;

	if(SetFilePointer(file, fsize.LowPart, &fsize.HighPart, FILE_BEGIN) == INVALID_SET_FILE_POINTER && GetLastError() != NO_ERROR)
		return -1;
#else
	if(lseek(file, offset, SEEK_SET) < 0)
		return -1;
#endif /* WIN32 */

	return 0;
}

int pdtp__file_read(pdtp_file_t file, void *buf, unsigned int nbytes)
{
#ifdef WIN32
	DWORD ret;
	
	if(!ReadFile(file, buf, nbytes, &ret, 0))
		return -1;
#else
	int ret;
	
	if((ret = read(file, buf, nbytes)) < 0)
		return -1;
#endif /* WIN32 */

	return ret;
}

int pdtp__file_write(pdtp_file_t file, void *buf, unsigned int nbytes)
{
#ifdef WIN32
	DWORD ret;

	if(!WriteFile(file, buf, nbytes, &ret, 0))
		return -1;
#else
	int ret;

	if((ret = write(file, buf, nbytes)) < 0)
		return -1;
#endif /* WIN32 */

	return ret;
}
