/*
   hasher.c - Abstract hash cipher interface
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote 
 products derived from this software without specific prior written 
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT 
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"

#include "alloc.h"
#include "hasher.h"
#include "md5.h"
#include "sha1.h"

struct pdtp_hasher {
	unsigned digest_size;
	void *handle;
	
	void (*destroy)(void *);
	void (*zero)(void *);
	void (*add)(void *, void *, unsigned);
	void (*get_digest)(void *, void *);
};

struct hash_cipher {
	unsigned digest_size;			/* Size of digest in bytes */
	void *(*create)();			/* Create a hasher */
	void (*destroy)(void *);		/* Destory a hasher */
	void (*zero)(void *);			/* Zero hasher contents */
	void (*add)(void *, void *, unsigned);	/* Hash data set */
	void (*get_digest)(void *, void *);	/* Get digest */
};

const struct hash_cipher ciphers[2] = {
	/* MD5 */
	{ 16, 
	(void *(*)())pdtp__md5_create, 
	(void (*)(void *))pdtp__md5_destroy, 
	(void (*)(void *))pdtp__md5_zero, 
	(void (*)(void *, void *, unsigned))pdtp__md5_add_data, 
	(void (*)(void *, void *))pdtp__md5_get_digest },
	
	/* SHA1 */
	{ 20,
	(void *(*)())pdtp__sha1_create,
	(void (*)(void *))pdtp__sha1_destroy,
	(void (*)(void *))pdtp__sha1_zero,
	(void (*)(void *, void *, unsigned))pdtp__sha1_add_data,
	(void (*)(void *, void *))pdtp__sha1_get_digest }
};

pdtp_hasher_t pdtp__hasher_create(pdtp_hash_t algorithm)
{
	pdtp_hasher_t hasher = NEW(pdtp_hasher);

	hasher->digest_size = ciphers[algorithm].digest_size;
	hasher->handle = ciphers[algorithm].create();

	hasher->destroy = ciphers[algorithm].destroy;
	hasher->zero = ciphers[algorithm].zero;
	hasher->add = ciphers[algorithm].add;
	hasher->get_digest = ciphers[algorithm].get_digest;

	return hasher;
}

void pdtp__hasher_destroy(pdtp_hasher_t hasher)
{
	hasher->destroy(hasher->handle);
	pdtp__free(hasher);
}

unsigned pdtp__hasher_digest_size(pdtp_hasher_t hasher)
{
	return hasher->digest_size;
}

void pdtp__hasher_zero(pdtp_hasher_t hasher)
{
	hasher->zero(hasher->handle);
}

void pdtp__hasher_add_data(pdtp_hasher_t hasher, void *buf, unsigned len)
{
	hasher->add(hasher->handle, buf, len);
}

void pdtp__hasher_get_digest(pdtp_hasher_t hasher, void *digest)
{
	hasher->get_digest(hasher->handle, digest);
}
