/*
   listener.c - Listener socket control functions
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote
 products derived from this software without specific prior written
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "pdtp.h"

#include <string.h>

#include "alloc.h"
#include "collection.h"
#include "listener.h"
#include "machdep.h"
#include "notifier.h"
#include "socketops.h"
#include "threadops.h"

typedef struct pdtp_listener_list_node {
	pdtp_notifier_t notifier;
	struct pdtp_listener_list_node *next;
} *pdtp_listener_list_node_t;

typedef struct pdtp_port_list_node {
	uint16_t port;
	pdtp_collection_t filter_collection;
	struct pdtp_port_list_node *next;
} *pdtp_port_list_node_t;

typedef struct pdtp_listener_params {
	int sock;
	pdtp_notifier_t notifier;

	int (*function)(pdtp_multiplexer_t, int sock, uint16_t opcode, uint32_t addr, uint16_t port, uint32_t handle, uint32_t piece_index, void *arg);
	void *arg;

	uint16_t opcode;
	uint32_t addr;
	uint16_t port;
	uint32_t handle, piece;
} *pdtp_listener_params_t;

struct pdtp_listener {
	uint32_t addr;
	uint32_t handle;
};

/* XXX Locking in this module is grossly incomplete.  A thorough evaluation
   of all this code for possible locking problems is in order */

static pdtp_sock_t listen_sock = -1;
static int active_listeners = 0;
static pdtp_listener_list_node_t listener_list = 0;
static pdtp_mutex_t listener_list_lock = 0;

static pdtp_notifier_pool_t notifier_pool = 0;
static pdtp_mutex_t notifier_pool_lock = 0;

static pdtp_collection_t listener_filters = 0;
static pdtp_mutex_t listener_filters_lock = 0;

int pdtp_listener_set_port(int port)
{
	if(!listener_list_lock)
		listener_list_lock = pdtp__mutex_create();

	if(!notifier_pool_lock)
		notifier_pool_lock = pdtp__mutex_create();

	if(!notifier_pool)
		notifier_pool = pdtp__notifier_pool_create();

	if(!listener_filters_lock)
		listener_filters_lock = pdtp__mutex_create();
	
	if(!listener_filters)
		listener_filters = pdtp__collection_create();

	if(listen_sock != -1)
		pdtp__socket_close(listen_sock);

	if((listen_sock = pdtp__ipv4_tcp_listener_create(0, port)) < 0) {
		listen_sock = -1;
		return -1;
	}

	return 0;
}

void pdtp_listener_destroy()
{
	if(listen_sock == -1)
		return;
	
	pdtp__socket_close(listen_sock);
	listen_sock = -1;
}

static pdtp_port_list_node_t pdtp__port_list_node_create(uint16_t port)
{
	pdtp_port_list_node_t node = NEW(pdtp_port_list_node);

	node->port = port;
	node->filter_collection = pdtp__collection_create();
	node->next = 0;

	return node;
}

pdtp_notifier_t pdtp__listener_notifier_create(uint32_t addr, uint16_t port, uint32_t handle)
{
	pdtp_notifier_t notifier;
	pdtp_port_list_node_t port_list, n, t;

	if(!notifier_pool_lock || listen_sock == -1)
		return 0;

	pdtp__mutex_lock(notifier_pool_lock);
	notifier = pdtp__notifier_get(notifier_pool);
	pdtp__mutex_unlock(notifier_pool_lock);

	pdtp__mutex_lock(listener_filters_lock);
	if(!(port_list = pdtp__collection_remove(listener_filters, addr)))
		n = port_list = pdtp__port_list_node_create(port);
	else if(port_list->port > port) {
		t = port_list;
		n = port_list = pdtp__port_list_node_create(port);
		port_list->next = t;
	} else {
		for(n = port_list; n->next && n->next->port < port; n = n->next);

		if(n->port != port) {
			t = n->next;
			n->next = pdtp__port_list_node_create(port);
			n = n->next;
			n->next = t;
		}
	}

	pdtp__collection_insert(n->filter_collection, handle, notifier);
	pdtp__collection_insert(listener_filters, addr, port_list);
	pdtp__mutex_unlock(listener_filters_lock);

	return notifier;
}

pdtp_notifier_t pdtp__listener_notifier_lookup(uint32_t addr, uint16_t port, uint32_t handle)
{
	return 0;
}

/*static void pdtp__listener_notifier_destroy(pdtp_notifier_t notifier)
{
}
*/

static int pdtp__process_transfer_request(pdtp_multiplexer_t mplx, int sock, pdtp_listener_params_t params)
{
	int ret;
	pdtp_notifier_t notifier;

	if(!(notifier = pdtp__listener_notifier_lookup(params->addr, params->port, params->handle)))
		goto err;
	
	if(notifier == params->notifier) {
		ret = params->function(mplx, sock, params->opcode, params->addr, params->port, params->handle, params->piece, params->arg);
		/* XXX Free params? */
		pdtp__free(params);
		return 1;
	}

	pdtp__notifier_send_message(notifier, params);

	return 1;

err:
	pdtp__socket_close(sock);

	/* XXX Free params? */
	pdtp__free(params);
	return 1;
}

static int pdtp__process_incoming_header(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg)
{
	uint8_t *bptr = (uint8_t *)buffer;
	uint32_t v32; 
	uint16_t v16;
	pdtp_listener_params_t params = (pdtp_listener_params_t)arg;
	
	if(bytes != 18)
		goto err;

	/* Check protocol version - only accept version 1 at this point */
	memcpy(&v16, bptr, 2);
	if(v16 != 1)
		goto err;

	/* Extract opcode */
	memcpy(&v16, bptr + 2, 2);
	params->opcode = htons(v16);

	/* Extract server address */
	memcpy(&v32, bptr + 4, 4);
	params->addr = htonl(v32);

	/* Extract server port */
	memcpy(&v16, bptr + 8, 2);
	params->port = htons(v16);

	/* Extract file handle */
	memcpy(&v32, bptr + 10, 4);
	params->handle = htonl(v32);

	/* Extract piece ID */
	memcpy(&v32, bptr + 14, 4);
	params->piece = htonl(v32);
	
	return pdtp__process_transfer_request(mplx, sock, params);

err:
	pdtp__socket_close(sock);

	/* XXX Free params? */
	pdtp__free(params);
	return 1;
}

static int pdtp__read_incoming_header(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg)
{
	pdtp__nonblock_read(mplx, sock, 18, pdtp__process_incoming_header, arg);

	return 1;
}

static int pdtp__check_incoming_magic(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg)
{
	const char *PTPP_MAGIC = "PTPP";

	if(bytes < 0)
		goto err;

	if(memcmp(buffer, PTPP_MAGIC, 4))
		goto err;

	pdtp__nonblock_write(mplx, sock, "PTPP\0\1\0\1", 8, pdtp__read_incoming_header, arg);

	return 1;

err:
	pdtp__socket_close(sock);

	/* XXX Free params? */
	pdtp__free(arg);
	return 1;
}

static int pdtp__ipv4_listener_handler(pdtp_multiplexer_t mplx, pdtp_sock_t listener, void *arg)
{
	pdtp_sock_t sock;
	uint32_t addr;
	
	if((sock = pdtp__ipv4_listener_accept(listener, &addr)) < 0) {
		/* XXX Free params? */
		pdtp__free(arg);
		return 1;
	}

	pdtp__nonblock_read(mplx, sock, 4, pdtp__check_incoming_magic, arg);

	return 1;
}

static int pdtp__listener_notifier_handler(pdtp_multiplexer_t mplx, pdtp_notifier_t notifier, void *arg)
{
	int ret;
	pdtp_listener_params_t params = (pdtp_listener_params_t)arg;
	
	ret = params->function(mplx, params->sock, params->opcode, params->addr, params->port, params->handle, params->piece, params->arg);

	/* XXX Free params? */
	pdtp__free(params);
	
	return ret;
}

void pdtp__multiplexer_add_ipv4_listener(pdtp_multiplexer_t mplx, pdtp_notifier_t notifier, int (*function)(pdtp_multiplexer_t, int fd, uint16_t opcode, uint32_t addr, uint16_t port, uint32_t handle, uint32_t piece_index, void *arg), void *arg)
{
	pdtp_listener_list_node_t node;
	pdtp_listener_params_t params = NEW(pdtp_listener_params);

	params->function = function;
	params->arg = arg;
	params->sock = listen_sock;
	params->notifier = notifier;

	pdtp__mutex_lock(listener_list_lock);

	if(!active_listeners++) {
		pdtp__mutex_unlock(listener_list_lock);
		pdtp__multiplexer_add_socket(mplx, listen_sock, MPLX_READ, pdtp__ipv4_listener_handler, params, 0);

		return;
		
	} 

	node = NEW(pdtp_listener_list_node);
	node->next = listener_list;
	listener_list = node;

	node->notifier = notifier;
	pdtp__mutex_unlock(listener_list_lock);
	
	pdtp__multiplexer_add_notifier(mplx, notifier, pdtp__listener_notifier_handler, params);
}

void *pdtp__multiplexer_remove_listener(pdtp_multiplexer_t mplx, pdtp_notifier_t notifier)
{
	return 0;
}
