/*
   multiplexer.c - Platform independent socket multiplexer mechanisms
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote
 products derived from this software without specific prior written
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "sockets.h"

#include <string.h>

#include "alloc.h"
#include "debug.h"
#include "multiplexer.h"
#include "notifier.h"

struct socket_event_entry {
	int sock;
	int (*function)(pdtp_multiplexer_t, int, void *);
	void *arg;
};

struct notifier_event_entry {
	pdtp_notifier_t notifier;
	int (*function)(pdtp_multiplexer_t, pdtp_notifier_t, void *);
	void *arg;
};

struct event_entry {
	enum { SOCKET_ENTRY, NOTIFIER_ENTRY } type;
	union {
		struct socket_event_entry sock_data;
		struct notifier_event_entry notifier_data;
	} data;
};

typedef struct pending_entry_node {
	struct event_entry data;
	int ts_sec;
	int ts_msec;

	struct pending_entry_node *next;
} *pending_entry_node_t;

typedef struct pending_entry_list {
	struct pending_entry_node *head, *tail, *spare;
	int spare_count;
} *pending_entry_list_t;

/* Delayed insertion code, temporarily commented out to suppress warnings */

#if 0
static pending_entry_list_t pelist_create(void)
{
	pending_entry_list_t list = NEW(pending_entry_list);

	list->head = list->tail = list->spare = 0;
	list->spare_count = 0;

	return list;
}

static void pelist_destroy(pending_entry_list_t list)
{
	pending_entry_node_t node;

	while((node = list->head)) {
		list->head = node->next;

		pdtp__free(node);
	}

	while((node = list->spare)) {
		list->spare = node->next;

		pdtp__free(node);
	}

	pdtp__free(list);
}

static void pelist_add(pending_entry_list_t list, struct event_entry *entry, int delay)
{
	struct timeval tv;

	gettimeofday(&tv, 0);
}
#endif

/* Win32 implementation.  Currently limited to 64 event objects due to
   constructions of the platform.  64 event objects should be enough
   for each transfer, so this restriction should not pose a problem */
#ifdef WIN32
struct pdtp_multiplexer {
	WSAEVENT event_table[WSA_MAXIMUM_WAIT_EVENTS];
	struct event_entry event_data[WSA_MAXIMUM_WAIT_EVENTS];
	int event_table_count;
};

pdtp_multiplexer_t pdtp__multiplexer_create(void)
{
	pdtp_multiplexer_t mplx = NEW(pdtp_multiplexer);

	mplx->event_table_count = 0;

	return mplx;
}

void pdtp__multiplexer_destroy(pdtp_multiplexer_t mplx)
{
	int i;

	for(i = 0; i < mplx->event_table_count; i++)
		CloseHandle(mplx->event_table[i]);

	pdtp__free(mplx);
}

void pdtp__multiplexer_add_socket(pdtp_multiplexer_t mplx, int sock, pdtp_event_t event_type, int (*function)(pdtp_multiplexer_t, int sock, void *arg), void *arg, int msec_delay)
{
	WSAEVENT event = WSACreateEvent();

	switch(event_type) {
		case MPLX_READ:
			WSAEventSelect(sock, event, FD_READ | FD_CLOSE);
			break;
		case MPLX_WRITE:
			WSAEventSelect(sock, event, FD_WRITE | FD_CLOSE);
			break;
	}

	mplx->event_table[mplx->event_table_count] = event;

	mplx->event_data[mplx->event_table_count].type = SOCKET_ENTRY;
	mplx->event_data[mplx->event_table_count].data.sock_data.sock = sock;
	mplx->event_data[mplx->event_table_count].data.sock_data.function = function;
	mplx->event_data[mplx->event_table_count].data.sock_data.arg = arg;

	mplx->event_table_count++;
}

void *pdtp__multiplexer_remove_socket(pdtp_multiplexer_t mplx, int sock)
{
	int i;
	void *ret;

	for(i = 0; i < mplx->event_table_count && (mplx->event_data[i].type == NOTIFIER_ENTRY || mplx->event_data[i].data.sock_data.sock != sock); i++);

	if(i == mplx->event_table_count)
		return 0;

	ret = mplx->event_data[i].data.sock_data.arg;

	CloseHandle(mplx->event_table[i]);

	if(i + 1 < mplx->event_table_count) {
		mplx->event_table[i] = mplx->event_table[mplx->event_table_count - 1];
		memcpy(&mplx->event_data[i], &mplx->event_data[mplx->event_table_count - 1], sizeof(struct event_entry));
	}

	mplx->event_table_count--;

	return ret;
}

void pdtp__multiplexer_add_notifier(pdtp_multiplexer_t mplx, pdtp_notifier_t notifier, int (*function)(pdtp_multiplexer_t, pdtp_notifier_t, void *), void *arg)
{
	mplx->event_table[mplx->event_table_count] = (WSAEVENT)pdtp__notifier_multiplex_object(notifier);

	mplx->event_data[mplx->event_table_count].type = NOTIFIER_ENTRY;
	mplx->event_data[mplx->event_table_count].data.notifier_data.notifier = notifier;
	mplx->event_data[mplx->event_table_count].data.notifier_data.function = function;
	mplx->event_data[mplx->event_table_count].data.notifier_data.arg = arg;

	mplx->event_table_count++;
}

void *pdtp__multiplexer_remove_notifier(pdtp_multiplexer_t mplx, pdtp_notifier_t notifier)
{
	int i;
	void *ret;

	for(i = 0; i < mplx->event_table_count && (mplx->event_data[i].type == SOCKET_ENTRY || mplx->event_data[i].data.notifier_data.notifier != notifier); i++);

	if(i == mplx->event_table_count)
		return 0;

	ret = mplx->event_data[i].data.notifier_data.arg;

	CloseHandle(mplx->event_table[i]);

	if(i + 1 < mplx->event_table_count) {
		mplx->event_table[i] = mplx->event_table[mplx->event_table_count - 1];
		memcpy(&mplx->event_data[i], &mplx->event_data[mplx->event_table_count - 1], sizeof(struct event_entry));
	}

	mplx->event_table_count--;

	return ret;
}

/* XXX This implementation doesn't support delayed addition and simply
   exits the function call in the event of an error.  The loop structure
   will need to change significantly to support the former, and in
   the process the latter can be remedied */
void pdtp__multiplexer_run(pdtp_multiplexer_t mplx)
{
	int i;

	while((i = WSAWaitForMultipleEvents(mplx->event_table_count, mplx->event_table, FALSE, INFINITE, FALSE)) != WSA_WAIT_FAILED) {
		if(mplx->event_data[i].type == SOCKET_ENTRY) {
			if(!mplx->event_data[i].data.sock_data.function(mplx, mplx->event_data[i].data.sock_data.sock, mplx->event_data[i].data.sock_data.arg))
				return;
		} else {
			if(!mplx->event_data[i].data.notifier_data.function(mplx, mplx->event_data[i].data.notifier_data.notifier, mplx->event_data[i].data.notifier_data.arg))
				return;
		}
	}
}
#endif

#ifdef HAVE_POLL
#include <sys/poll.h>

/* Initial size of the event table */
#define INITIAL_TABLE_SIZE      16

struct pdtp_multiplexer {
	struct pollfd *fd_table;
	struct event_entry *fd_data;
	int fd_table_count, fd_table_size;
	int current_index;
};

pdtp_multiplexer_t pdtp__multiplexer_create(void)
{
	pdtp_multiplexer_t mplx = NEW(pdtp_multiplexer);

#ifndef NDEBUG
	debug("Creating new multiplexer: %p", mplx);
#endif

	mplx->fd_table = (struct pollfd *)pdtp__alloc(INITIAL_TABLE_SIZE * sizeof(struct pollfd));
	mplx->fd_data = (struct event_entry *)pdtp__alloc(INITIAL_TABLE_SIZE * sizeof(struct event_entry));
	mplx->fd_table_count = 0;
	mplx->fd_table_size = INITIAL_TABLE_SIZE;

	return mplx;
}

void pdtp__multiplexer_destroy(pdtp_multiplexer_t mplx)
{
#ifndef NDEBUG
	debug("Destroying multiplexer: %p", mplx);
#endif

	pdtp__free(mplx->fd_table);
	pdtp__free(mplx->fd_data);
	pdtp__free(mplx);
}

void pdtp__multiplexer_add_socket(pdtp_multiplexer_t mplx, int sock, pdtp_event_t event_type, int (*function)(pdtp_multiplexer_t, int sock, void *arg), void *arg, int msec_delay)
{
	void *ptr;

	mplx->fd_table[mplx->fd_table_count].fd = sock;

	switch(event_type) { 
		case MPLX_READ: 
			mplx->fd_table[mplx->fd_table_count].events = POLLIN;
			break;
		case MPLX_WRITE:
			mplx->fd_table[mplx->fd_table_count].events = POLLOUT;
			break;
	}

	mplx->fd_data[mplx->fd_table_count].type = SOCKET_ENTRY;
	mplx->fd_data[mplx->fd_table_count].data.sock_data.sock = sock;
	mplx->fd_data[mplx->fd_table_count].data.sock_data.function = function;
	mplx->fd_data[mplx->fd_table_count].data.sock_data.arg = arg;

	mplx->fd_table_count++;

	/* Increase the size of the file descriptor table if it is full */
	if(mplx->fd_table_count == mplx->fd_table_size) {
		mplx->fd_table_size *= 2;

		/* XXX Perhaps in the future a realloc() wrapper could be
		   added within the library framework to make this process
		   more efficient */
		ptr = mplx->fd_table;
		mplx->fd_table = pdtp__alloc(mplx->fd_table_size * sizeof(struct pollfd));
		memcpy(mplx->fd_table, ptr, sizeof(struct pollfd) * mplx->fd_table_count);
		pdtp__free(ptr);

		ptr = mplx->fd_data;
		mplx->fd_data = pdtp__alloc(mplx->fd_table_size * sizeof(struct event_entry));
		memcpy(mplx->fd_data, ptr, sizeof(struct event_entry) * mplx->fd_table_count);
		pdtp__free(ptr);
	}
}

void *pdtp__multiplexer_remove_socket(pdtp_multiplexer_t mplx, int sock)
{
	int i;
	void *ret, *ptr;

	for(i = 0; i < mplx->fd_table_count && (mplx->fd_data[i].type != SOCKET_ENTRY || mplx->fd_data[i].data.sock_data.sock != sock); i++);

	if(i == mplx->fd_table_count)
		return 0;

	ret = mplx->fd_data[i].data.sock_data.arg;

	if(i + 1 < mplx->fd_table_count) {
		memcpy(&mplx->fd_table[i], &mplx->fd_table[mplx->fd_table_count - 1], sizeof(struct pollfd));
		memcpy(&mplx->fd_data[i], &mplx->fd_data[mplx->fd_table_count - 1], sizeof(struct event_entry));
	}

	mplx->fd_table_count--;

	/* If only 1/4 of the file descriptor table is full, shrink it */
	if(mplx->fd_table_count <= mplx->fd_table_size / 4 && mplx->fd_table_size / 2 >= INITIAL_TABLE_SIZE) {
		mplx->fd_table_size /= 2;

		/* XXX Perhaps in the future a realloc() wrapper could be
		   added within the library framework to make this process
		   more efficient */
		ptr = mplx->fd_table;
		mplx->fd_table = pdtp__alloc(mplx->fd_table_size * sizeof(struct pollfd));
		memcpy(mplx->fd_table, ptr, sizeof(struct pollfd) * mplx->fd_table_count);
		pdtp__free(ptr);

		ptr = mplx->fd_data;
		mplx->fd_data = pdtp__alloc(mplx->fd_table_size * sizeof(struct event_entry));
		memcpy(mplx->fd_data, ptr, sizeof(struct event_entry) * mplx->fd_table_count);
		pdtp__free(ptr);
	}

	return ret;
}

void pdtp__multiplexer_add_notifier(pdtp_multiplexer_t mplx, pdtp_notifier_t notifier, int (*function)(pdtp_multiplexer_t, pdtp_notifier_t, void *), void *arg)
{
	void *ptr;

	mplx->fd_table[mplx->fd_table_count].fd = pdtp__notifier_multiplex_object(notifier);
	mplx->fd_table[mplx->fd_table_count].events = POLLIN;

	mplx->fd_data[mplx->fd_table_count].type = NOTIFIER_ENTRY;
	mplx->fd_data[mplx->fd_table_count].data.notifier_data.notifier = notifier;
	mplx->fd_data[mplx->fd_table_count].data.notifier_data.function = function;
	mplx->fd_data[mplx->fd_table_count].data.notifier_data.arg = arg;

	mplx->fd_table_count++;

	/* Increase the size of the file descriptor table if it is full */
	if(mplx->fd_table_count == mplx->fd_table_size) {
		mplx->fd_table_size *= 2;

		/* XXX Perhaps in the future a realloc() wrapper could be
		   added within the library framework to make this process
		   more efficient */
		ptr = mplx->fd_table;
		mplx->fd_table = pdtp__alloc(mplx->fd_table_size * sizeof(struct pollfd));
		memcpy(mplx->fd_table, ptr, sizeof(struct pollfd) * mplx->fd_table_count);
		pdtp__free(ptr);

		ptr = mplx->fd_data;
		mplx->fd_data = pdtp__alloc(mplx->fd_table_size * sizeof(struct event_entry));
		memcpy(mplx->fd_data, ptr, sizeof(struct event_entry) * mplx->fd_table_count);
		pdtp__free(ptr);
	}
}

void *pdtp__multiplexer_remove_notifier(pdtp_multiplexer_t mplx, pdtp_notifier_t notifier)
{
	int i;
	void *ret, *ptr;

	for(i = 0; i < mplx->fd_table_count && (mplx->fd_data[i].type != NOTIFIER_ENTRY || mplx->fd_data[i].data.notifier_data.notifier != notifier); i++);

	if(i == mplx->fd_table_count)
		return 0;

	ret = mplx->fd_data[i].data.notifier_data.arg;

	if(i + 1 < mplx->fd_table_count) {
		memcpy(&mplx->fd_table[i], &mplx->fd_table[mplx->fd_table_count - 1], sizeof(struct pollfd));
		memcpy(&mplx->fd_data[i], &mplx->fd_data[mplx->fd_table_count - 1], sizeof(struct event_entry));
	}

	mplx->fd_table_count--;

	/* If only 1/4 of the file descriptor table is full, shrink it */
	if(mplx->fd_table_count <= mplx->fd_table_size / 4 && mplx->fd_table_size / 2 >= INITIAL_TABLE_SIZE) {
		mplx->fd_table_size /= 2;

		/* XXX Perhaps in the future a realloc() wrapper could be
		   added within the library framework to make this process
		   more efficient */
		ptr = mplx->fd_table;
		mplx->fd_table = pdtp__alloc(mplx->fd_table_size * sizeof(struct pollfd));
		memcpy(mplx->fd_table, ptr, sizeof(struct pollfd) * mplx->fd_table_count);
		pdtp__free(ptr);

		ptr = mplx->fd_data;
		mplx->fd_data = pdtp__alloc(mplx->fd_table_size * sizeof(struct event_entry));
		memcpy(mplx->fd_data, ptr, sizeof(struct event_entry) * mplx->fd_table_count);
		pdtp__free(ptr);
	}

	return ret;
}

/* XXX This implementation doesn't support delayed addition and simply
   exits the function call in the event of an error.  The loop structure
   will need to change significantly to support the former, and in
   the process the latter can be remedied 
   */
void pdtp__multiplexer_run(pdtp_multiplexer_t mplx)
{
	int remaining_events;

	while((remaining_events = poll(mplx->fd_table, mplx->fd_table_count, -1)) > 0) {
		for(mplx->current_index = 0; remaining_events && mplx->current_index < mplx->fd_table_count; mplx->current_index++) {
			if(mplx->fd_table[mplx->current_index].revents) {
				remaining_events--;

				if(mplx->fd_data[mplx->current_index].type == SOCKET_ENTRY) {
					if(!mplx->fd_data[mplx->current_index].data.sock_data.function(mplx, mplx->fd_data[mplx->current_index].data.sock_data.sock, mplx->fd_data[mplx->current_index].data.sock_data.arg))
						return;
				} else {
					if(!mplx->fd_data[mplx->current_index].data.notifier_data.function(mplx, mplx->fd_data[mplx->current_index].data.notifier_data.notifier, mplx->fd_data[mplx->current_index].data.notifier_data.arg))                                return;
				}
			}
		}
	}
}
#endif

/* A select()-based implementation for non-Win32 systems which also lack the 
   poll() system call, namely MacOS X 10.0 through 10.2.  This deficiency
   was addressed in 10.3, which added poll() support.  */
#if !defined(WIN32) && !defined(HAVE_POLL)

/* Initial size of the event table */
#define INITIAL_TABLE_SIZE	16

struct pdtp_multiplexer {
	fd_set read_table, write_table;
	struct event_entry *fd_data;
	int fd_table_count, fd_table_size;
	int current_index, maxfd;
};

pdtp_multiplexer_t pdtp__multiplexer_create(void)
{
	pdtp_multiplexer_t mplx = NEW(pdtp_multiplexer);

	FD_ZERO(&mplx->read_table);
	FD_ZERO(&mplx->write_table);
	mplx->maxfd = -1;

	mplx->fd_data = (struct event_entry *)pdtp__alloc(INITIAL_TABLE_SIZE * sizeof(struct event_entry));
	mplx->fd_table_count = 0;
	mplx->fd_table_size = INITIAL_TABLE_SIZE;

	return mplx;
}

void pdtp__multiplexer_destroy(pdtp_multiplexer_t mplx)
{
	pdtp__free(mplx->fd_data);
	pdtp__free(mplx);
}

void pdtp__multiplexer_add_socket(pdtp_multiplexer_t mplx, int sock, pdtp_event_t event_type, int (*function)(pdtp_multiplexer_t, int sock, void *arg), void *arg, int msec_delay)
{
	struct event_entry *old_data;

	if(event_type == MPLX_READ)
		FD_SET(sock, &mplx->read_table);
	else
		FD_SET(sock, &mplx->write_table);

	if(sock > mplx->maxfd)
		mplx->maxfd = sock;

	mplx->fd_data[mplx->fd_table_count].type = SOCKET_ENTRY;
	mplx->fd_data[mplx->fd_table_count].data.sock_data.sock = sock;
	mplx->fd_data[mplx->fd_table_count].data.sock_data.function = function;
	mplx->fd_data[mplx->fd_table_count].data.sock_data.arg = arg;

	mplx->fd_table_count++;

	if(mplx->fd_table_count == mplx->fd_table_size) {
		mplx->fd_table_size *= 2;

		/* XXX Perhaps in the future a realloc() wrapper could be
		   added within the library framework to make this process
		   more efficient */
		old_data = mplx->fd_data;
		mplx->fd_data = pdtp__alloc(mplx->fd_table_size * sizeof(struct event_entry));
		memcpy(mplx->fd_data, old_data, sizeof(struct event_entry) * mplx->fd_table_count);
		pdtp__free(old_data);
	}
}

void *pdtp__multiplexer_remove_socket(pdtp_multiplexer_t mplx, int sock)
{
	struct event_entry *old_data;
	void *ret;
	int i;

	FD_CLR(sock, &mplx->read_table);
	FD_CLR(sock, &mplx->write_table);

	for(i = 0; i < mplx->fd_table_count && (mplx->fd_data[i].type != SOCKET_ENTRY || mplx->fd_data[i].data.sock_data.sock != sock); i++);

	if(i == mplx->fd_table_count)
		return 0;

	ret = mplx->fd_data[i].data.sock_data.arg;

	if(i + 1 < mplx->fd_table_count) 
		memcpy(&mplx->fd_data[i], &mplx->fd_data[mplx->fd_table_count - 1], sizeof(struct event_entry));

	mplx->fd_table_count--;

	/* If only 1/4 of the file descriptor table is full, shrink it */
	if(mplx->fd_table_count <= mplx->fd_table_size / 4 && mplx->fd_table_size / 2 >= INITIAL_TABLE_SIZE) {
		mplx->fd_table_size /= 2;

		/* XXX Perhaps in the future a realloc() wrapper could be
		   added within the library framework to make this process
		   more efficient */
		old_data = mplx->fd_data;
		mplx->fd_data = pdtp__alloc(mplx->fd_table_size * sizeof(struct event_entry));
		memcpy(mplx->fd_data, old_data, sizeof(struct event_entry) * mplx->fd_table_count);
		pdtp__free(old_data);
	}

	return ret;
}

void pdtp__multiplexer_add_notifier(pdtp_multiplexer_t mplx, pdtp_notifier_t notifier, int (*function)(pdtp_multiplexer_t, pdtp_notifier_t, void *), void *arg)
{
	void *ptr;
	int fd = pdtp__notifier_multiplex_object(notifier);

        FD_SET(fd, &mplx->read_table);

        if(fd > mplx->maxfd)
                mplx->maxfd = fd;

        mplx->fd_data[mplx->fd_table_count].type = NOTIFIER_ENTRY;
        mplx->fd_data[mplx->fd_table_count].data.notifier_data.notifier = notifier;
        mplx->fd_data[mplx->fd_table_count].data.notifier_data.function = function;
        mplx->fd_data[mplx->fd_table_count].data.notifier_data.arg = arg;

        mplx->fd_table_count++;

        /* Increase the size of the file descriptor table if it is full */
        if(mplx->fd_table_count == mplx->fd_table_size) {
                mplx->fd_table_size *= 2;

                /* XXX Perhaps in the future a realloc() wrapper could be
                   added within the library framework to make this process
                   more efficient */
                ptr = mplx->fd_data;
                mplx->fd_data = pdtp__alloc(mplx->fd_table_size * sizeof(struct event_entry));
                memcpy(mplx->fd_data, ptr, sizeof(struct event_entry) * mplx->fd_table_count);
                pdtp__free(ptr);
        }
}

void *pdtp__multiplexer_remove_notifier(pdtp_multiplexer_t mplx, pdtp_notifier_t notifier)
{
	int i;
	void *ret, *ptr;

	for(i = 0; i < mplx->fd_table_count && (mplx->fd_data[i].type != NOTIFIER_ENTRY || mplx->fd_data[i].data.notifier_data.notifier != notifier); i++);

	if(i == mplx->fd_table_count)
		return 0;

	ret = mplx->fd_data[i].data.notifier_data.arg;

	if(i + 1 < mplx->fd_table_count)
		memcpy(&mplx->fd_data[i], &mplx->fd_data[mplx->fd_table_count - 1], sizeof(struct event_entry));

	mplx->fd_table_count--;

	/* If only 1/4 of the file descriptor table is full, shrink it */
	if(mplx->fd_table_count <= mplx->fd_table_size / 4 && mplx->fd_table_size / 2 >= INITIAL_TABLE_SIZE) {
		mplx->fd_table_size /= 2;

		/* XXX Perhaps in the future a realloc() wrapper could be
		   added within the library framework to make this process
		   more efficient */
		ptr = mplx->fd_data;
		mplx->fd_data = pdtp__alloc(mplx->fd_table_size * sizeof(struct event_entry));
		memcpy(mplx->fd_data, ptr, sizeof(struct event_entry) * mplx->fd_table_count);
		pdtp__free(ptr);
	}

	return ret;
}

void pdtp__multiplexer_run(pdtp_multiplexer_t mplx)
{
	int fd, remaining_events;
	fd_set read_table, write_table;

	for(;;) {
		/* Copy the read/write fd sets from our prebuilt ones.
		   This should be faster than the relatively branch heavy
		   code and the modulus calculations required to build
		   these from our event_data structures. */
		memcpy(&read_table, &mplx->read_table, sizeof(fd_set));
		memcpy(&write_table, &mplx->write_table, sizeof(fd_set));

		if((remaining_events = select(mplx->maxfd + 1, &read_table, &write_table, 0, 0)) <= 0)
			break;

		for(mplx->current_index = 0; remaining_events && mplx->current_index < mplx->fd_table_count; mplx->current_index++) {
			if(mplx->fd_data[mplx->current_index].type == SOCKET_ENTRY)
				fd = mplx->fd_data[mplx->current_index].data.sock_data.sock;
			else
				fd = pdtp__notifier_multiplex_object(mplx->fd_data[mplx->current_index].data.notifier_data.notifier);

			if(FD_ISSET(fd, &read_table) || FD_ISSET(fd, &write_table)) {
				if(mplx->fd_data[mplx->current_index].type == SOCKET_ENTRY) { 
					if(!mplx->fd_data[mplx->current_index].data.sock_data.function(mplx, mplx->fd_data[mplx->current_index].data.sock_data.sock, mplx->fd_data[mplx->current_index].data.sock_data.arg)) 
						return;
				} else {
					if(!mplx->fd_data[mplx->current_index].data.notifier_data.function(mplx, mplx->fd_data[mplx->current_index].data.notifier_data.notifier, mplx->fd_data[mplx->current_index].data.notifier_data.arg))            
						return;
				}
			}
		}
	}
}
#endif
