/*
   notifier.c - Portable, multiplexable interthread messaging mechanism
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote 
 products derived from this software without specific prior written 
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT 
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "sockets.h"

#include "alloc.h"
#include "notifier.h"
#include "threadops.h"

/* Maximum number of spare notifiers allowed in a given notifier pool */
#define MAX_NOTIFIER_POOL_SIZE	8

/* The notifier structure.  On Win32 we can multiplex event objects with 
   sockets, allowing us to multiplex the pdtp_notifier_t object with a
   socket.  Unfortunately, POSIX does not provide event objects with
   file descriptors, so we are forced to use pipes for interthread 
   communication as they can be multiplexed with the select() or poll()
   system calls (and will bring the thread out of the respective system
   call when we write to them, which is where the destination thread
   for the message will be blocking */
struct pdtp_notifier {
#ifdef WIN32
	void *message;
	pdtp_mutex_t mutex;
	pdtp_cond_t cond;
#else
	int pipe[2];
#endif /* WIN32 */
};

/* To mitigate the overhead of the system calls required to create or
   destroy notifier objects, we store notifier objects in a shared
   resource pool for each connection. */
typedef struct pdtp_notifier_pool_frame {
	pdtp_notifier_t notifier;
	struct pdtp_notifier_pool_frame *next;
} *pdtp_notifier_pool_frame_t;

struct pdtp_notifier_pool {
	pdtp_notifier_pool_frame_t root;
	int count;
};

static pdtp_notifier_t notifier_create(void)
{
	pdtp_notifier_t notifier;
	
	notifier = NEW(pdtp_notifier);
	
#ifdef WIN32
	notifier->message = 0;
	notifier->mutex = pdtp__mutex_create();
	notifier->cond = pdtp__cond_create();
#else
	if(pipe(notifier->pipe) < 0) {
		pdtp__free(notifier);
		return 0;
	}
#endif /* WIN32 */

	return notifier;
}

static void notifier_destroy(pdtp_notifier_t notifier)
{
#ifdef WIN32
	pdtp__mutex_destroy(notifier->mutex);
	pdtp__cond_destroy(notifier->cond);
#else
	close(notifier->pipe[0]);
	close(notifier->pipe[1]);
#endif

	pdtp__free(notifier);
}

pdtp_notifier_pool_t pdtp__notifier_pool_create(void)
{
	pdtp_notifier_pool_t pool = NEW(pdtp_notifier_pool);

	pool->root = 0;
	pool->count = 0;
	
	return pool;
}

void pdtp__notifier_pool_destroy(pdtp_notifier_pool_t pool)
{
	pdtp_notifier_pool_frame_t frame;
	
	while(pool->root) {
		frame = pool->root;
		pool->root = frame->next;
		
		notifier_destroy(frame->notifier);
		pdtp__free(frame);
	}

	pdtp__free(pool);
}

pdtp_notifier_t pdtp__notifier_get(pdtp_notifier_pool_t pool)
{
	pdtp_notifier_t notifier;
	pdtp_notifier_pool_frame_t frame;

	if(!pool->root)
		return notifier_create();

	frame = pool->root;
	pool->root = frame->next;
	
	notifier = frame->notifier;
	pdtp__free(frame);

	pool->count--;

	return notifier;
}

void pdtp__notifier_return(pdtp_notifier_pool_t pool, pdtp_notifier_t notifier)
{
	pdtp_notifier_pool_frame_t frame;
	
	if(pool->count >= MAX_NOTIFIER_POOL_SIZE) {
		notifier_destroy(notifier);
		return;
	}

	frame = NEW(pdtp_notifier_pool_frame);
	
	frame->notifier = notifier;
	frame->next = pool->root;
	pool->root = frame;

	pool->count++;
}

void pdtp__notifier_send_message(pdtp_notifier_t notifier, void *message)
{
#ifdef WIN32
	pdtp__mutex_lock(notifier->mutex);
	notifier->message = message;
	pdtp__cond_signal(notifier->cond);
	pdtp__mutex_unlock(notifier->mutex);
#else
	write(notifier->pipe[1], &message, sizeof(void *));
#endif /* WIN32 */
}

void *pdtp__notifier_recv_message(pdtp_notifier_t notifier)
{
	void *ret;
#ifdef WIN32
	pdtp__mutex_lock(notifier->mutex);
	pdtp__cond_wait(notifier->cond, notifier->mutex);
	ret = notifier->message;
	pdtp__mutex_unlock(notifier->mutex);
#else
	if(read(notifier->pipe[0], &ret, sizeof(void *)) < 0)
		return 0;
#endif /* WIN32 */

	return ret;
}

#ifdef WIN32
WSAEVENT pdtp__notifier_multiplex_object(pdtp_notifier_t notifier)
{
	return (WSAEVENT)notifier->cond;
}
#else
int pdtp__notifier_multiplex_object(pdtp_notifier_t notifier)
{
	return notifier->pipe[0];
}
#endif /* WIN32 */
