/*
   This code provides example source for implementing a unithreaded PDTP
   client using libpdtp (although its primary use is as a test driver)
   It was written by Tony Arcieri in 2003-04.  No copyright is claimed.
   This code is in the public domain; do with it as you see fit.
*/

#include "machdep.h"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#ifdef WIN32
#include <windows.h>
#endif

#include "pdtp.h"

#define PDTP_DEFAULT_PORT	4043

pdtp_connection_t conn = 0;
char *conn_host = 0;
char *cwd = 0;

static void do_open(char *arg);
static void do_close(char *arg);
static void do_ls(char *arg);
static void do_cd(char *arg);
static void do_cwd(char *arg);
static void do_pwd(char *arg);
static void do_get(char *arg);
static void do_port(char *arg);
static void do_help(char *arg);

struct command_list_entry {
	char *command;
	void (*function)(char *);
};

struct command_list_entry command_list[] = {
	{ "open", do_open },
	{ "close", do_close },
	{ "ls", do_ls },
	{ "cd", do_cd },
	{ "cwd", do_cwd },
	{ "pwd", do_pwd },
	{ "get", do_get },
	{ "port", do_port },
	{ "help", do_help },
	{ "quit", 0 },
	{ "exit", 0 },
	{ "bye", 0 },
	{ 0, 0 }
};

void print_error()
{
	char *error = pdtp_get_last_error();

	if(!error) {
		fputs("An unknown error occured\n", stderr);
		return;
	}

	fprintf(stderr, "Error: %s\n", error);
}

static char *input_line(const char *prompt)
{
	char buf[512], *t;

	fputs(prompt, stdout);
	fflush(stdout);

	if(!fgets(buf, 512, stdin))
		return 0;

	if((t = strpbrk(buf, "\r\n")))
		*t = '\0';

	return strdup(buf);
}

static void do_open(char *arg)
{
	int port = PDTP_DEFAULT_PORT;
	char *t;

	if(conn) {
		puts("Already connected.  Use 'close' first.");
		return;
	}

	if(!arg || !*arg) {
		if(!(conn_host = input_line("(to) ")))
			return;
	} else
		conn_host = strdup(arg);

	if((t = strpbrk(conn_host, " :"))) {
		*t++ = '\0';
		port = atoi(t);
	}

	if(!(conn = pdtp_connect(conn_host, port))) {
		free(conn_host);
		print_error();
	} else
		printf("Connected to %s\n", conn_host);

	cwd = strdup("/");
}

static void do_close(char *arg)
{
	if(!conn) {
		puts("Not connected.");
		return;
	}

	pdtp_close(conn);
	conn = 0;

	printf("Connection to %s closed.\n", conn_host);
	free(conn_host);
	free(cwd);
}

static void do_ls(char *arg)
{
	pdtp_directory_t dir;
	pdtp_direntry_t entry;
	time_t t;
	char entry_name[PDTP_PATH_MAX];

	if(!conn) {
		puts("Not connected.");
		return;
	}

	printf("Listing directory: %s\n", cwd);

	if(!(dir = pdtp_get_directory_listing(conn, cwd, PDTP_STORE))) {
		print_error();
		return;
	}

	while(!pdtp_get_direntry(dir, &entry)) {
		t = (entry.ts_year - 1970) * 31556926 + entry.ts_ysec;

		switch(entry.type) {
			case PDTP_DIR:
				strncpy(entry_name, entry.name, PDTP_PATH_MAX);
				strncat(entry_name, "/", PDTP_PATH_MAX);
				printf("DIR          %-32s %s", entry_name, ctime(&t));
				break;
			case PDTP_FILE:
				printf("%-12d %-32s %s", (int)entry.length, entry.name, ctime(&t));
				break;
		}
	}

	pdtp_directory_close(dir);
}

static void do_cd(char *arg)
{
	int l;
	char *t;

	if(!conn) { 
		puts("Not connected."); 
		return; 
	}

	if(!arg || !*arg) { 
		if(!(arg = input_line("(remote-directory) "))) 
			return;
	}

	if(!strcmp(arg, "."))
		return;

	if(!strcmp(arg, "..")) {
		l = strlen(cwd);
		if(cwd[l - 1] == '/')
			cwd[l - 1] = '\0';

		if((t = strrchr(cwd, '/')))
			t[1] = '\0';
	} else {
		l = strlen(cwd) + strlen(arg) + 2;

		t = (char *)malloc(l);
		strcpy(t, cwd);
		if(cwd[strlen(cwd) - 1] != '/') 
			strcat(t, "/");
		strcat(t, arg);

		free(cwd);
		cwd = t;
	}

	printf("Working directory is now: %s\n", cwd);
}

static void do_cwd(char *arg)
{
	if(!conn) {
		puts("Not connected.");
		return;
	}

	if(!arg || !*arg) {
		if(!(arg = input_line("(remote-directory) ")))
			return;

		free(cwd);
		cwd = arg;
	} else {
		free(cwd);
		cwd = strdup(arg);
	}

	printf("Working directory set to: %s\n", cwd);
}

static void do_pwd(char *arg)
{
	if(!conn) {
		puts("Not connected.");
		return;
	}

	printf("Working directory is currently: %s\n", cwd);
}

static void do_get(char *arg)
{
	char path[PDTP_PATH_MAX], *error;
	pdtp_download_t dl;
	pdtp_file_t file;

	if(!conn) {
		puts("Not connected.");
		return;
	}

	strncpy(path, cwd, PDTP_PATH_MAX);
	if(path[strlen(path) - 1] != '/')
		strncat(path, "/", PDTP_PATH_MAX);
	strncat(path, arg, PDTP_PATH_MAX);

	if(!(dl = pdtp_download_create(conn, path))) {
		error = pdtp_get_last_error();

		if(error) 
			fprintf(stderr, "Download initialization error: %s\n", error);		
		else
			fputs("Unknown download initialization error.\n", stderr);
		return; 
	}

	if(pdtp_fopen(arg, &file) < 0) {
		perror(arg);

		pdtp_download_destroy(dl);
		return;
	}
	
	pdtp_download_run(dl, file, 0, 0);
	pdtp_download_destroy(dl);

	pdtp_fclose(file);
}

static void do_port(char *arg)
{
	int port = atoi(arg);

	if(port < 1) {
		puts("Disabling listener port");
		pdtp_listener_destroy();

		return;
	}
	
	printf("Setting listener port to %d... ", port);
	fflush(stdout);

	if(pdtp_listener_set_port(port) < 0)
		puts("failed.");
	else
		puts("ok.");
}

static void do_help(char *arg)
{
	int i = 0;
	if(!conn) {
		puts("Not connected.");
	}
    else {
		printf("connected to %s\n",conn_host);
    }
	printf("Working directory set to: %s\n", cwd);

    puts("available commands:");
	while(command_list[i].command) {
		puts(command_list[i].command);
        i++;
    }
}


static void command_exec(char *command, char *arg)
{
	int i = 0;

	while(command_list[i].command) {
		if(strcmp(command_list[i].command, command)) {
			i++;
			continue;
		}

		if(!command_list[i].function) {
			free(command);

			if(conn) {
				pdtp_close(conn);
				free(conn_host);
				free(cwd);
			}

			exit(0);
		}

		command_list[i].function(arg);
		return;
	}

	printf("Unknown command: %s\n", command);
}

int main(int argc, char **argv)
{
	char *command, *arg;

#ifdef WIN32
	WSADATA data;

	if(WSAStartup(MAKEWORD(2, 0), &data)) {
		fputs("Error: Couldn't initialize Winsock", stderr);
		exit(1);
	}
#endif

	if(argc > 1)
		do_open(argv[1]);

	while((command = input_line("pdtp> "))) {
		if((arg = strchr(command, ' ')))
			*arg++ = '\0';
		else
			arg = "";

		command_exec(command, arg);
		free(command);
	}

	exit(0);
}
