/*
   pdtp.h - Interface of the PDTP client library, libpdtp
   Last revised 1 January 2004

   Copyright (C)2003-04 Anthony Arcieri.  All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote 
 products derived from this software without specific prior written 
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT 
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
/**
 * @mainpage Interface of the PDTP client library, libpdtp
 */
/**
 * @file pdtp.h
 * @brief Interface of the PDTP client library, libpdtp
 * @remark
 *************************
 * IMPORTANT INFORMATION *
 *************************

 libpdtp is still under *heavy* development.  Many of the functions in
 the library are still only partially implemented or remain completely
 unimplemented.  Furthermore, the interface is subject to change.

 Users of libpdtp on Win32 should include pdtp.h AFTER including windows.h
 to ensure that fpos_t is used in place of off_t.

 All functions which return handle types will return NULL in the event of
 an error.  Be certain to check for NULL returns from these functions.
 An interface for returning error messages has not yet been defined,
 but will be in the future.
 */

#ifndef PDTP_H
#define PDTP_H

#include <sys/types.h>
#include <stdio.h>

/** Maximum path length as defined by the PDTP specification */
#define PDTP_PATH_MAX	1024

/* Platform specific types */
#ifdef WIN32
typedef HANDLE pdtp_file_t;
typedef __int64 pdtp_filelen_t;
#else
/** file handle */
typedef int pdtp_file_t;
/** length of a file */
typedef off_t pdtp_filelen_t;
#endif

/***************************
 * Handle Type Definitions *
 ***************************/

/** pdtp_connection_t represents a connection to a PDTP server */
typedef struct pdtp_connection *pdtp_connection_t;

/** pdtp_proxy_t represents a connection to a PDTP proxy */
typedef struct pdtp_proxy_t *pdtp_proxy_t;

/** pdtp_directory_t is a handle to a directory listing */
typedef struct pdtp_directory *pdtp_directory_t;

/** pdtp_download_t represents a file download */
typedef struct pdtp_download *pdtp_download_t;

/*****************************
 * Listener Socket Interface *
 *****************************/

/**
 * @defgroup Listen Listener Socket Interface 
 * @{
 */

/** pdtp_listener_set_port() creates a listener socket if one isn't already
   active, and changes the port being listened otherwise.  If a connection is
   created without an active listener socket, the connection will operate
   in passive mode only.  The port of the listener socket may be changed
   by a subsequent call to pdtp_listener_set_port() */
int pdtp_listener_set_port(int port);

/** pdtp_listener_destroy() destroys the listener socket and all associated
   structures and threads.  It will cause any active transfers to revert to
   passive mode as well. */
void pdtp_listener_destroy(void);

/** @} */
/************************
 * Connection Interface *
 ************************/
/**
 * @defgroup Connection Connection Interfaces
 * @{
 */


/** pdtp_connect() creates a connection to the specified server */
pdtp_connection_t pdtp_connect(const char *address, int port);

/** pdtp_close() closes the socket associated with the given connection handle
   and frees internal data associated with the handle */
void pdtp_close(pdtp_connection_t);

/**
    @defgroup ConnectProxy  Connecting via a Proxy 
    @ingroup Connection
   Hosts located behind a firewall can use a PDTP proxy both to establish
   connections and to accept active mode piece transfers.  Users needing
   to use a proxy to make outgoing connections should use the following
   functions to connect to a server 
    @{
    */
/** connect to the proxy */
pdtp_proxy_t pdtp_proxy_connect(const char *address, int port);

/** connect through a proxy */
pdtp_connection_t pdtp_connect_through_proxy(const char *address, int port, pdtp_proxy_t);
/** close proxy */
void pdtp_proxy_close(pdtp_proxy_t);


/** Hosts which can make outgoing connections but need a proxy to make
   incoming connections can configure the connection to use a proxy
   host for incoming connections using the functions below.  A proxy
   must be set before a download is initiated for it to have an effect */
void pdtp_set_proxy(pdtp_connection_t, const char *address, int port);
/** unset the proxy */
void pdtp_unset_proxy(pdtp_connection_t);
/** @} */

/** @} */
/*******************************
 * Directory Listing Interface *
 *******************************/

/**
 * @defgroup dirlist Directory Listing
 * @{
 */


/** Directory entry types */
typedef enum {
	PDTP_FILE, /**< FILE Type */
	PDTP_DIR   /**<  Directory Type */
} pdtp_entry_type_t;

/** Directory entry information structure. */
typedef struct {
	/** Year component of timestamp */
	int ts_year;

	/** Number of seconds since year component in timestamp */
	int ts_ysec;

	/** Entry name */
	char name[PDTP_PATH_MAX];

	/** Length of file in bytes, 0 for subdirectories */
	pdtp_filelen_t length;

	/** Entry type, one of PDTP_FILE or PDTP_DIR */
	pdtp_entry_type_t type;
} pdtp_direntry_t;

/** Two methods of listing directories are supported.  PDTP_FETCH allows
   the directory listing to be processed as it is read from the server.
   PDTP_STORE reads the entire directory listing from the server first,
   then returns a handle to the complete result. */
typedef enum {
	PDTP_FETCH,	/**< Fetch the directory listing incrementally */
	PDTP_STORE	/**< Store the results before returning */
} pdtp_read_mode_t;

/** This returns a listing of the specified directory */
pdtp_directory_t pdtp_get_directory_listing(pdtp_connection_t, const char *path, pdtp_read_mode_t mode);

/** This closes a directory listing handle */
void pdtp_directory_close(pdtp_directory_t);

/** This rewinds the directory handle to the first entry in the directory */
void pdtp_directory_rewind(pdtp_directory_t);

/** This function is used when reading file or subdirectory entries.  Each
   successive call will populate the referenced PDTP_DirEntry structure
   with the next entry in the directory.  Returns 0 on success, 1 when
   the end of directory is reached, and -1 on error.  A buffer for the
   "name" member is allocated with each call using the internal allocator
   (typically malloc) and must be freed by the programmer using the library */
int pdtp_get_direntry(pdtp_directory_t, pdtp_direntry_t *entry);

/** @} */
/***************************
 * File Transfer Interface *
 ***************************/

/**
 * @defgroup FT File Transfer Interface
 * @{
 */


/** Function table for event notification.  This table is passed to the
   pdtp_download_run() function.  The last argument of pdtp_download_run()
   is a void pointer which will be passed as the last argument to any
   function in this table. */
typedef struct {
	/** This is invoked when we actually begin the download */
	void (*download_begin)(int pieces_remaining, int pieces_total, void *ptr);

	/** This is invoked when the download has been successfully completed */
	void (*download_complete)(void *ptr);

	/** This is invoked if a download fails. If this occurs,
	   pdtp_download_run() will also exit */ 
	void (*download_fail)(const char *message, void *ptr);

	/** This is invoked when we attempt to connect to a peer */
	void (*peer_connect)(const char *address, int port, void *ptr);

	/** This is invoked when we successfully connect to a peer */
	void (*peer_connect_complete)(const char *address, int port, void *ptr);

	/** This is invoked if a connection to a peer fails */
	void (*peer_connect_fail)(const char *address, int port, const char *errmsg, void *ptr);

	/** This is invoked when we disconnect from a peer */
	void (*peer_disconnect)(char *address, int port, void *ptr);

	/** This is invoked when a peer connects to us */
	void (*incoming_connect)(char *address, int port, void *ptr);

	/** This is invoked when a peer disconnects from us */
	void (*incoming_disconnect)(char *address, int port, void *ptr);

	/** This is invoked when we begin to download a piece */
	void (*piece_download)(char *address, int port, int piece, int length, void *ptr);

	/** This is invoked when we begin to upload a piece */
	void (*piece_upload)(char *address, int port, int piece, int length, void *ptr);

	/** This is invoked when a piece download completes */
	void (*piece_download_complete)(char *address, int port, int piece, void *ptr);

	/** This is invoked when a piece upload completes */
	void (*piece_upload_complete)(char *address, int port, int piece, void *ptr);

	/** This is invoked whenever we download a portion of a piece */
	void (*piece_chunk_download)(char *address, int port, int piece, int length, void *ptr);

	/** This is invoked whenever we upload a portion of a piece */
	void (*piece_chunk_upload)(char *address, int port, int piece, int length, void *ptr);

	/** This is invoked when we begin checksumming a piece */
	void (*piece_checksum)(int piece, int length, void *ptr);

	/** This is invoked if a piece checksum matches */
	void (*piece_checksum_complete)(int piece, void *ptr);

	/** This is invoked if a piece checksum mismatches */
	void (*piece_checksum_fail)(int piece, void *ptr);

	/** This is invoked whenever we MD5 checksum a piece */
	void (*piece_chunk_scan)(int piece, int length, void *ptr);
} pdtp_callback_table_t;

/** This initiates a new download */
pdtp_download_t pdtp_download_create(pdtp_connection_t, const char *path);

/** This will free the information associated with a download */
void pdtp_download_destroy(pdtp_download_t);

/** The following are options which may be set before a download starts
   or from another thread during an active download */
typedef enum {
	/** Setting PDTP_USE_BACKGROUND_THREAD to true will detach the transfer
	   into a background thread rather than executing the download event
	   loop in the context of the current thread.  This means that all
	   callback functions defined in a pdtp_callback_table_t will be
	   executed in the context of this background thread rather than
	   in the context of the thread which calls pdtp_download_run().
	   If you use this option, make sure all callbacks are thread safe
	   for whatever threads implementation libpdtp has been linked with.
	   Zero is false, non-zero is true.  The default value is 0 */
	PDTP_USE_BACKGROUND_THREAD,
	
	/** Setting PDTP_EXIT_AFTER_COMPLETE to true will cause 
	   pdtp_download_run() to exit immediately after a download completes
	   rather than continue to serve pieces to other peers.  
	   Zero is false, non-zero is true.  The default value is 0. */
	PDTP_EXIT_AFTER_COMPLETE,

	/** Maximum upload speed in bytes per second.  Negative values
	   indicate no limit.  The default value is -1 */
	PDTP_MAX_UPLOAD_SPEED,

	/** Maximum download speed in bytes per second.  Negative values
	   indicate no limit.  The default value is -1 */
	PDTP_MAX_DOWNLOAD_SPEED
} pdtp_download_option_t;

/** This is the function to set one of the above options. */
void pdtp_download_set_option(pdtp_download_t, pdtp_download_option_t, int value);

/** This function begins the download.  It will only return if the download
   fails, if PDTP_EXIT_AFTER_COMPLETE is true, or if pdtp_download_stop
   is invoked from another thread.  No other threads should be manipulating
   the given file while this function is active.  This function does not 
   handle closing the given file pointer, that is left up to the programmer 
   using the library. */
void pdtp_download_run(pdtp_download_t, pdtp_file_t, pdtp_callback_table_t *, void *ptr);

/** This function may be invoked from another thread to gracefully terminate
   an active transfer */
void pdtp_download_stop(pdtp_download_t);

/* @} */
/******************
 * File Interface * 
 ******************/
/**
 * @defgroup File File Interface
 * @{
 */


/** pdtp_fopen() will create the specified file if it doesn't exist, or simply
   open it if it does.  The file is opened for reading and writing, which
   is required in order to pass it to pdtp_download_run */
int pdtp_fopen(char *filename, pdtp_file_t *file);

/** pdtp_fclose() closes the given pdtp_file_t */
void pdtp_fclose(pdtp_file_t file);

/** @} */
/****************************
 * Error Handling Interface *
 ****************************/
/**
 * @defgroup Error Error Handling Interface
 * @{
 */


/** pdtp_get_last_error() will return a pointer to a buffer containing the
   last error message.  This function is implemented using thread-specific 
   data and is thread safe. */
char *pdtp_get_last_error(void);

/** pdtp_set_last_error() will overwrite the value in the current thread's
   error message buffer with the given message.  The error message is copied
   from the given string, and may be altered/freed after this function is
   invoked without altering the value stored internally. */
void pdtp_set_last_error(const char *message);

/** @} */
/*******************************
 * Memory Management Interface *
 *******************************/
/**
 * @defgroup memmgmt Memory Management Interface
 * @{
 */


/** This function lets you override the internal out-of-memory handler for
   libpdtp.  By default this function will call perror() for the given
   allocator function name and then call abort().  libpdtp expects
   this function will never return. */
void pdtp_oom_handler_set(void (*function)(const char *allocator_name));

/** This allows you to override the internal allocation function used by
   libpdtp, which is normally a wrapper around the system malloc().  libpdtp
   will not check for a NULL return from the specified allocator function.
   The default behavior will be to call abort() if malloc() returns NULL. */
void pdtp_alloc_set(void *(*function)(int nbytes));

/** This allows you to override the internal free function, normally the
   system's free() */
void pdtp_free_set(void (*function)(void *ptr));

/** @} */
#endif
