/*
   piecelist.c - Piece list generation and management code
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote 
 products derived from this software without specific prior written 
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT 
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "pdtp.h"

#include <string.h>

#include "alloc.h"
#include "fileops.h"
#include "md5.h"
#include "piecelist.h"
#include "txid.h"

#define HASH_BLOCK_SIZE 16384
#define CHECKSUM_SIZE	16

typedef struct pdtp_piecelistnode { 
	int start, end; 
	struct pdtp_piecelistnode *next;
} *pdtp_piecelistnode_t;

struct pdtp_piecelist {
	pdtp_piecelistnode_t head, tail;
	int count;
};

static void piecelist_append(pdtp_piecelist_t list, int i)
{
	if(!list->head) {
		list->head = list->tail = NEW(pdtp_piecelistnode);
		list->head->start = i;
		list->head->end = i;
		list->head->next = 0;
		list->count = 1;

		return;
	}

	/* This function only supports appending entries to a list */
	if(list->tail->end >= i)
		return;

	if(list->tail->end == i - 1) {
		list->tail->end++;
		return;
	}

	list->tail->next = NEW(pdtp_piecelistnode);
	list->tail = list->tail->next;

	list->tail->start = i;
	list->tail->end = i;
	list->tail->next = 0;
	list->count++;
}

/* This function handles both truncating the file to the proper length and
   MD5 checksumming the file in order to build a piece list */
pdtp_piecelist_t pdtp__piecelist_create(pdtp_file_t file, pdtp_filelen_t length, uint32_t piece_length, uint8_t *hashes)
{
	unsigned i, j, n, b;
	unsigned piece_count, total_piece_count, block_count, block_length;
	uint8_t buffer[HASH_BLOCK_SIZE];
	unsigned char checksum[CHECKSUM_SIZE];
	pdtp_filelen_t orig_length;
	pdtp_piecelist_t list;
	pdtp_md5_t md5;

	/* Get the current length of the given file */
	if(pdtp__file_length(file, &orig_length) < 0)
		return 0;

	/* If the length mismatches our destination length, truncate the file */
	if(orig_length != length) {
		if(pdtp__file_truncate(file, length) < 0)
			return 0;
	}

	/* Return the file pointer to the beginning of the file */
	if(pdtp__file_seek(file, 0) < 0)
		return 0;

	md5 = pdtp__md5_create();

	/* We only checksum the original data, rather than the full truncated
	   space.  This prevents us from checksumming known bad data */
	piece_count = (unsigned)(orig_length / piece_length);

	if(length == orig_length && orig_length % piece_length)
		piece_count++;

	list = NEW(pdtp_piecelist);
	list->head = list->tail = 0;
	list->count = 0;

	for(i = 0; i < piece_count; i++) {
		pdtp__md5_zero(md5);
		block_count = piece_length / HASH_BLOCK_SIZE;

		if(piece_length % HASH_BLOCK_SIZE)
			block_count++;

		block_length = piece_length;

		for(j = 0; j < block_count; j++) {
			b = block_length > HASH_BLOCK_SIZE ? HASH_BLOCK_SIZE : block_length;

			if((n = pdtp__file_read(file, buffer, b)) < 0) {
				/* XXX Better error handling is in order */
				return 0;
			}

			pdtp__md5_add_data(md5, buffer, n);
			block_length -= n;
		}

		pdtp__md5_get_digest(md5, checksum);

		if(memcmp(checksum, hashes + i * CHECKSUM_SIZE, CHECKSUM_SIZE))
			piecelist_append(list, i);
	}

	/* Here we calculate the total number of pieces, and we will 
	   automatically mark any pieces that don't exist in the original
	   file as needing to be transferred */
	total_piece_count = (unsigned)(length / piece_length);

	if(length % piece_length)
		total_piece_count++;

	/* Here we append all pieces that weren't in the original file to the
	   piece list as they obviously need to be retransferred. */
	for(i = piece_count; i < total_piece_count; i++)
		piecelist_append(list, i); /* XXX Do this more efficiently */

	return list;
}

void pdtp__piecelist_destroy(pdtp_piecelist_t list)
{
	pdtp_piecelistnode_t n;

	while(list->head) {
		n = list->head;
		list->head = n->next;

		pdtp__free(n);
	}

	pdtp__free(list);
}

void pdtp__piecelist_add_to_transaction(pdtp_transaction_out_t txn, pdtp_piecelist_t list)
{
	int i = 0; /* Debugging */
	pdtp_piecelistnode_t cur;
	uint32_t *buffer;

	buffer = (uint32_t *)pdtp__alloc(list->count * 8);

	for(cur = list->head; cur; cur = cur->next) {
		buffer[i++] = cur->start;
		buffer[i++] = cur->end;
	}

	pdtp__transaction_add_object(txn, OBJ_PIECELIST, buffer, list->count * 8);
	pdtp__free(buffer);
}
