/* 
   SHA1.c - NIST Secure Hash Algorithm 

   Modified for PDTP use in 2004 by Anthony Arcieri
   Heavily modified by Uwe Hollerbach <uh@alumni.caltech edu> 
   from Peter C. Gutmann's implementation as found in 
   Applied Cryptography by Bruce Schneier 

   This code is in the public domain 
 */

#include "machdep.h"

#include <string.h>

#include "alloc.h"
#include "sha1.h"

#define SHA_BLOCKSIZE           64
#define SHA_DIGESTSIZE          20

struct pdtp_sha1 {
	uint32_t digest[5];             /* message digest */
	uint32_t count_lo, count_hi;    /* 64-bit bit count */
	uint8_t data[SHA_BLOCKSIZE];    /* SHA1 data buffer */
	int local;                      /* unprocessed amount in data */
};

/* SHA1 f()-functions */

#define f1(x,y,z)((x & y) | (~x & z))
#define f2(x,y,z)(x ^ y ^ z)
#define f3(x,y,z)((x & y) | (x & z) | (y & z))
#define f4(x,y,z)(x ^ y ^ z)

/* SHA1 constants */

#define CONST1	0x5a827999L
#define CONST2	0x6ed9eba1L
#define CONST3	0x8f1bbcdcL
#define CONST4	0xca62c1d6L

/* truncate to 32 bits -- should be a null op on 32-bit machines */

#define T32(x)((x) & 0xffffffffL)

/* 32-bit rotate */

#define R32(x,n)T32(((x << n) | (x >> (32 - n))))

/* the generic case, for when the overall rotation is not unraveled */

#define FG(n)\
T = T32(R32(A,5) + f##n(B,C,D) + E + *WP++ + CONST##n);\
	E = D; D = C; C = R32(B,30); B = A; A = T

	/* specific cases, for when the overall rotation is unraveled */

#define FA(n)\
	T = T32(R32(A,5) + f##n(B,C,D) + E + *WP++ + CONST##n); B = R32(B,30)

#define FB(n)\
	E = T32(R32(T,5) + f##n(A,B,C) + D + *WP++ + CONST##n); A = R32(A,30)

#define FC(n)\
	D = T32(R32(E,5) + f##n(T,A,B) + C + *WP++ + CONST##n); T = R32(T,30)

#define FD(n)\
	C = T32(R32(D,5) + f##n(E,T,A) + B + *WP++ + CONST##n); E = R32(E,30)

#define FE(n)\
	B = T32(R32(C,5) + f##n(D,E,T) + A + *WP++ + CONST##n); D = R32(D,30)

#define FT(n)\
	A = T32(R32(B,5) + f##n(C,D,E) + T + *WP++ + CONST##n); C = R32(C,30)

pdtp_sha1_t pdtp__sha1_create()
{
	pdtp_sha1_t sha = NEW(pdtp_sha1);

	pdtp__sha1_zero(sha);

	return sha;
}

void pdtp__sha1_destroy(pdtp_sha1_t sha)
{
	pdtp__free(sha);
}

/* do SHA1 transformation */
static void pdtp__sha1_transform(pdtp_sha1_t sha)
{
	int i;
	uint8_t *dp;
	uint32_t T, A, B, C, D, E, W[80], *WP;

	dp = sha->data;

#if HOST_BIGENDIAN
	for (i = 0; i < 16; ++i) {
		T = *((uint32_t *) dp);
		dp += 4;
		W[i] = T32(T);
	}
#else
	for (i = 0; i < 16; ++i) {
		T = *((uint32_t *) dp);
		dp += 4;
		W[i] =  ((T << 24) & 0xff000000) | ((T <<  8) & 0x00ff0000) |
			((T >>  8) & 0x0000ff00) | ((T >> 24) & 0x000000ff);
	}
#endif /* HOST_BIGENDIAN */

	for (i = 16; i < 80; ++i) {
		W[i] = W[i-3] ^ W[i-8] ^ W[i-14] ^ W[i-16];
		W[i] = R32(W[i], 1);
	}

	A = sha->digest[0];
	B = sha->digest[1];
	C = sha->digest[2];
	D = sha->digest[3];
	E = sha->digest[4];
	WP = W;

	FA(1); FB(1); FC(1); FD(1); FE(1); FT(1); FA(1); FB(1); FC(1); FD(1);
	FE(1); FT(1); FA(1); FB(1); FC(1); FD(1); FE(1); FT(1); FA(1); FB(1);
	FC(2); FD(2); FE(2); FT(2); FA(2); FB(2); FC(2); FD(2); FE(2); FT(2);
	FA(2); FB(2); FC(2); FD(2); FE(2); FT(2); FA(2); FB(2); FC(2); FD(2);
	FE(3); FT(3); FA(3); FB(3); FC(3); FD(3); FE(3); FT(3); FA(3); FB(3);
	FC(3); FD(3); FE(3); FT(3); FA(3); FB(3); FC(3); FD(3); FE(3); FT(3);
	FA(4); FB(4); FC(4); FD(4); FE(4); FT(4); FA(4); FB(4); FC(4); FD(4);
	FE(4); FT(4); FA(4); FB(4); FC(4); FD(4); FE(4); FT(4); FA(4); FB(4);
	sha->digest[0] = T32(sha->digest[0] + E);
	sha->digest[1] = T32(sha->digest[1] + T);
	sha->digest[2] = T32(sha->digest[2] + A);
	sha->digest[3] = T32(sha->digest[3] + B);
	sha->digest[4] = T32(sha->digest[4] + C);
}

/* initialize the SHA1 digest */
void pdtp__sha1_zero(pdtp_sha1_t sha)
{
	sha->digest[0] = 0x67452301L;
	sha->digest[1] = 0xefcdab89L;
	sha->digest[2] = 0x98badcfeL;
	sha->digest[3] = 0x10325476L;
	sha->digest[4] = 0xc3d2e1f0L;
	sha->count_lo = 0L;
	sha->count_hi = 0L;
	sha->local = 0;
}

/* update the SHA1 digest */
void pdtp__sha1_add_data(pdtp_sha1_t sha, void *data, unsigned count)
{
	unsigned i;
	uint32_t clo;
	uint8_t *buffer = (uint8_t *)data;

	clo = T32(sha->count_lo + ((uint32_t) count << 3));

	if (clo < sha->count_lo) {
		++sha->count_hi;
	}

	sha->count_lo = clo;
	sha->count_hi += (uint32_t) count >> 29;

	if (sha->local) {
		i = SHA_BLOCKSIZE - sha->local;
		if (i > count) {
			i = count;
		}
		memcpy(((uint8_t *) sha->data) + sha->local, buffer, i);
		count -= i;
		buffer += i;
		sha->local += i;
		if (sha->local == SHA_BLOCKSIZE) {
			pdtp__sha1_transform(sha);
		} else {
			return;
		}
	}

	while (count >= SHA_BLOCKSIZE) {
		memcpy(sha->data, buffer, SHA_BLOCKSIZE);
		buffer += SHA_BLOCKSIZE;
		count -= SHA_BLOCKSIZE;
		pdtp__sha1_transform(sha);
	}
	memcpy(sha->data, buffer, count);
	sha->local = count;
}

/* finish computing the SHA1 digest */
void pdtp__sha1_get_digest(pdtp_sha1_t sha, unsigned char digest[20])
{
	int count;
	uint32_t lo_bit_count, hi_bit_count;

	lo_bit_count = sha->count_lo;
	hi_bit_count = sha->count_hi;
	count = (int) ((lo_bit_count >> 3) & 0x3f);
	((uint8_t *) sha->data)[count++] = 0x80;

	if (count > SHA_BLOCKSIZE - 8) {
		memset(((uint8_t *) sha->data) + count, 0, SHA_BLOCKSIZE - count);
		pdtp__sha1_transform(sha);
		memset((uint8_t *) sha->data, 0, SHA_BLOCKSIZE - 8);
	} else {
		memset(((uint8_t *) sha->data) + count, 0, SHA_BLOCKSIZE - 8 - count);
	}

	sha->data[56] = (hi_bit_count >> 24) & 0xff;
	sha->data[57] = (hi_bit_count >> 16) & 0xff;
	sha->data[58] = (hi_bit_count >>  8) & 0xff;
	sha->data[59] = (hi_bit_count >>  0) & 0xff;
	sha->data[60] = (lo_bit_count >> 24) & 0xff;
	sha->data[61] = (lo_bit_count >> 16) & 0xff;
	sha->data[62] = (lo_bit_count >>  8) & 0xff;
	sha->data[63] = (lo_bit_count >>  0) & 0xff;

	pdtp__sha1_transform(sha);

	digest[ 0] = (unsigned char) ((sha->digest[0] >> 24) & 0xff);
	digest[ 1] = (unsigned char) ((sha->digest[0] >> 16) & 0xff);
	digest[ 2] = (unsigned char) ((sha->digest[0] >>  8) & 0xff);
	digest[ 3] = (unsigned char) ((sha->digest[0]      ) & 0xff);
	digest[ 4] = (unsigned char) ((sha->digest[1] >> 24) & 0xff);
	digest[ 5] = (unsigned char) ((sha->digest[1] >> 16) & 0xff);
	digest[ 6] = (unsigned char) ((sha->digest[1] >>  8) & 0xff);
	digest[ 7] = (unsigned char) ((sha->digest[1]      ) & 0xff);
	digest[ 8] = (unsigned char) ((sha->digest[2] >> 24) & 0xff);
	digest[ 9] = (unsigned char) ((sha->digest[2] >> 16) & 0xff);
	digest[10] = (unsigned char) ((sha->digest[2] >>  8) & 0xff);
	digest[11] = (unsigned char) ((sha->digest[2]      ) & 0xff);
	digest[12] = (unsigned char) ((sha->digest[3] >> 24) & 0xff);
	digest[13] = (unsigned char) ((sha->digest[3] >> 16) & 0xff);
	digest[14] = (unsigned char) ((sha->digest[3] >>  8) & 0xff);
	digest[15] = (unsigned char) ((sha->digest[3]      ) & 0xff);
	digest[16] = (unsigned char) ((sha->digest[4] >> 24) & 0xff);
	digest[17] = (unsigned char) ((sha->digest[4] >> 16) & 0xff);
	digest[18] = (unsigned char) ((sha->digest[4] >>  8) & 0xff);
	digest[19] = (unsigned char) ((sha->digest[4]      ) & 0xff);
}
