/*
   socketops.c - Platform independent socket functions
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote 
 products derived from this software without specific prior written 
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT 
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "pdtp.h"

#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#ifndef WIN32
#include <errno.h>
#endif

#include "alloc.h"
#include "error.h"
#include "multiplexer.h"
#include "socketops.h"
#include "sockets.h"

/* Timeouts */
#define CONNECT_TIMEOUT		30
#define RESPONSE_TIMEOUT 	60

/* Connection backlog */
#define MAX_BACKLOG 256

struct pdtp_addr {
	struct sockaddr *saddr;
	socklen_t slen;
};

typedef struct nonblock_transfer_parameters {
	uint8_t *buffer;
	unsigned bytes_transferred, bytes_total;
	int (*function)(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg);
	void *arg;
} *nonblock_transfer_parameters_t;

pdtp_addr_t pdtp__ipv4_addr_create(uint32_t addr, uint16_t port)
{
	pdtp_addr_t ret = NEW(pdtp_addr);
	struct sockaddr_in *sin;

	ret->saddr = (struct sockaddr *)pdtp__alloc(sizeof(struct sockaddr_in));
	ret->slen = sizeof(struct sockaddr_in);
	memset(ret->saddr, 0, sizeof(struct sockaddr_in));
	sin = (struct sockaddr_in *)ret->saddr;

	sin->sin_family = AF_INET;
	sin->sin_port = htons((uint16_t)port);
	memcpy(&sin->sin_addr.s_addr, &addr, 4);

	return ret;
}

/* Hostname resolution function */
pdtp_addr_t pdtp__resolve(const char *hostname, int port)
{
#ifdef HAVE_GETADDRINFO 
	pdtp_addr_t ret;
	struct addrinfo hints, *res;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;

	if(getaddrinfo(hostname, 0, &hints, &res)) {
		pdtp__error_printf(ERROR_RESOLVE, hostname);
		return 0;
	}

	ret = NEW(pdtp_addr);
	ret->saddr = (struct sockaddr *)pdtp__alloc(sizeof(struct sockaddr_in));
	ret->slen = sizeof(struct sockaddr_in);

	memcpy(ret->saddr, res->ai_addr, sizeof(struct sockaddr_in));
	freeaddrinfo(res);

	((struct sockaddr_in *)ret->saddr)->sin_port = htons(port);
	pdtp__set_error_ptr(0);

	return ret;
#else
	/* XXX We should confirm whether or not gethostbyname() uses 
	   thread-specific data, and if not mutex lock the call (ouch) */
	pdtp_addr_t ret;
	struct sockaddr_in *sin;
	struct hostent *host;

	ret = NEW(pdtp_addr);
	ret->saddr = (struct sockaddr *)pdtp__alloc(sizeof(struct sockaddr_in));
	ret->slen = sizeof(struct sockaddr_in);
	memset(ret->saddr, 0, sizeof(struct sockaddr_in));
	sin = (struct sockaddr_in *)ret->saddr;

	sin->sin_family = AF_INET;
	sin->sin_port = htons((uint16_t)port);

	if((int)(sin->sin_addr.s_addr = inet_addr(hostname)) == -1) {
		if(!(host = gethostbyname(hostname)) || host->h_length != 4) {
			pdtp__error_printf(ERROR_RESOLVE, hostname);
			pdtp__free(ret);

			return 0;
		}

		memcpy(&sin->sin_addr, host->h_addr, 4);
#ifndef WIN32
		endhostent();
#endif
	}

	pdtp__set_error_ptr(0);

	return ret;
#endif /* HAVE_GETADDRINFO */
}

void pdtp__addr_destroy(pdtp_addr_t addr)
{
	pdtp__free(addr->saddr);
	pdtp__free(addr);
}

uint32_t pdtp__addr_ipv4(pdtp_addr_t addr)
{
	if(((struct sockaddr_in *)addr->saddr)->sin_family != AF_INET)
		return 0;

	return (uint32_t)((struct sockaddr_in *)addr->saddr)->sin_addr.s_addr;
}

uint16_t pdtp__addr_port(pdtp_addr_t addr)
{
	if(((struct sockaddr_in *)addr->saddr)->sin_family != AF_INET) 
		return 0;

	return ((struct sockaddr_in *)addr->saddr)->sin_port;
}

/* Connect function */
pdtp_sock_t pdtp__nonblock_connect(pdtp_addr_t addr)
{
	pdtp_sock_t sock;

	if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		goto err0;

	if(pdtp__set_blocking_mode(sock, PDTP_SOCKET_NONBLOCK) < 0)
		goto err1;

	if(connect(sock, addr->saddr, addr->slen) < 0) {
#ifdef WIN32
		if(WSAGetLastError() != WSAEWOULDBLOCK)
			goto err1;
#else
		if(errno != EINPROGRESS)
			goto err1;
#endif
	}

	pdtp__set_error_ptr(0);
	return sock;

err1:
	pdtp__socket_close(sock);
err0:

	pdtp_set_last_error(ERROR_CONNECT);
	return -1;
}

int pdtp__nonblock_connect_complete(pdtp_sock_t sock)
{
	int errval;
	socklen_t errlen;

	errlen = sizeof(int);

	if(getsockopt(sock, SOL_SOCKET, SO_ERROR, (char *)&errval, &errlen) < 0)
		goto err;

	if(errval)
		goto err;

	return 0;

err:
	pdtp_set_last_error(ERROR_CONNECT);
	return -1;
}

pdtp_sock_t pdtp__connect(pdtp_addr_t addr)
{
	pdtp_sock_t sock;
	fd_set fdset;
	struct timeval timeout;

	if((sock = pdtp__nonblock_connect(addr)) < 0)
		return -1;

	FD_ZERO(&fdset);
	FD_SET(sock, &fdset);

	timeout.tv_sec = CONNECT_TIMEOUT;
	timeout.tv_usec = 0;

#ifdef WIN32
	if(select((int)sock + 1, 0, 0, &fdset, &timeout) < 1)
		goto err;
#else
	if(select(sock + 1, 0, &fdset, 0, &timeout) < 1)
		goto err;
#endif

	if(pdtp__nonblock_connect_complete(sock) < 0)
		goto err;

	pdtp__set_blocking_mode(sock, PDTP_SOCKET_NONBLOCK);

	pdtp__set_error_ptr(0);
	return sock;

err:
	pdtp__socket_close(sock);
	pdtp_set_last_error(ERROR_CONNECT);
	return -1;
}

int pdtp__set_blocking_mode(pdtp_sock_t sock, pdtp_blockmode_t mode)
{
	unsigned long flags;

#ifdef WIN32
	switch(mode) {
		case PDTP_SOCKET_BLOCK:
			flags = 0;
			break;
		case PDTP_SOCKET_NONBLOCK:
			flags = 1;
			break;
	}

	if(ioctlsocket(sock, FIONBIO, &flags) == SOCKET_ERROR)
		return -1;
#else
	if((flags = fcntl(sock, F_GETFL)) < 0)
		return -1;

	switch(mode) {
		case PDTP_SOCKET_BLOCK:
			flags &= ~O_NONBLOCK;
			break;
		case PDTP_SOCKET_NONBLOCK:
			flags |= O_NONBLOCK;
			break;
	}

	if(fcntl(sock, F_SETFL, flags) < 0)
		return -1;
#endif /* WIN32 */

	return 0;
}

/* Atomic read function */
ssize_t pdtp__atomic_readsock(pdtp_sock_t sock, void *buf, size_t nbytes)
{
	return recv(sock, buf, nbytes, 0);
}

/* Read function with timeout */
ssize_t pdtp__timeout_readsock(pdtp_sock_t sock, void *buf, size_t nbytes)
{
	fd_set fdset;
	struct timeval tv;

	tv.tv_sec = RESPONSE_TIMEOUT;
	tv.tv_usec = 0;

	FD_ZERO(&fdset);
	FD_SET(sock, &fdset);

	if(select((int)sock + 1, &fdset, 0, 0, &tv) < 1) {
		pdtp_set_last_error(ERROR_READ_TIMEOUT);
		return -1;
	}

	return pdtp__atomic_readsock(sock, buf, nbytes);
}

/* Read function */
int pdtp__readsock(pdtp_sock_t sock, void *buf, size_t nbytes)
{
	ssize_t c;
	uint8_t *bptr = buf;

	while(nbytes > 0) {
		if((c = pdtp__timeout_readsock(sock, bptr, nbytes)) < 1)
			return -1;

		nbytes -= c;
		bptr += c;
	}

	return 0;
}

int pdtp__readsock_int16(pdtp_sock_t sock, uint16_t *value)
{
	if(pdtp__readsock(sock, value, 2) < 0)
		return -1;

	*value = ntohs(*value);

	return 0;
}

int pdtp__readsock_int32(pdtp_sock_t sock, uint32_t *value)
{
	if(pdtp__readsock(sock, value, 4) < 0)
		return -1;

	*value = ntohl(*value);

	return 0;
}

ssize_t pdtp__atomic_writesock(pdtp_sock_t sock, void *buf, size_t nbytes)
{
	return send(sock, buf, nbytes, 0);
}

ssize_t pdtp__timeout_writesock(pdtp_sock_t sock, void *buf, size_t nbytes)
{
	fd_set fdset;
	struct timeval tv;

	tv.tv_sec = RESPONSE_TIMEOUT;
	tv.tv_usec = 0;

	FD_ZERO(&fdset);
	FD_SET(sock, &fdset);

	if(select((int)sock + 1, 0, &fdset, 0, &tv) < 1)
		return -1;

	return send(sock, buf, nbytes, 0);
}

/* Write function.  This uses a loop to ensure the data is written completely,
   as the sockets involved will be set non-blocking */
int pdtp__writesock(pdtp_sock_t sock, void *buf, size_t nbytes)
{
	ssize_t c;
	uint8_t *bptr = buf;

	while(nbytes > 0) {
		if((c = pdtp__timeout_writesock(sock, bptr, nbytes)) < 1)
			return -1;

		nbytes -= c;
		bptr += c;
	}

	return 0;
}


int pdtp__writesock_int16(pdtp_sock_t sock, uint16_t value)
{
	uint16_t v = htons(value);

	return pdtp__writesock(sock, &v, 2);
}

int pdtp__writesock_int32(pdtp_sock_t sock, uint32_t value)
{
	uint32_t v = htonl(value);

	return pdtp__writesock(sock, &v, 4);
}

static int nonblock_read_handler(pdtp_multiplexer_t mplx, int sock, void *arg)
{
        int n, ret;
        nonblock_transfer_parameters_t params = (nonblock_transfer_parameters_t)arg;

        if((n = pdtp__atomic_readsock(sock, params->buffer + params->bytes_transferred, params->bytes_total - params->bytes_transferred)) < 0) {
                pdtp__multiplexer_remove_socket(mplx, sock);
                ret = params->function(mplx, sock, 0, -1, params->arg);

                pdtp__free(params->buffer);
                pdtp__free(params);

                return ret;
        }

        params->bytes_transferred += n;

        /* XXX Add throttling support here */
        if(params->bytes_transferred < params->bytes_total)
                return 1;

        pdtp__multiplexer_remove_socket(mplx, sock);
        ret = params->function(mplx, sock, params->buffer, params->bytes_total, params->arg);

        pdtp__free(params->buffer);
        pdtp__free(params);

        return ret;
}

void pdtp__nonblock_read(pdtp_multiplexer_t mplx, pdtp_sock_t sock, unsigned length, int (*completion_function)(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg), void *arg)
{
        nonblock_transfer_parameters_t params = NEW(nonblock_transfer_parameters);

        params->buffer = pdtp__alloc(length);
        params->bytes_transferred = 0;
        params->bytes_total = length;
        params->function = completion_function;
        params->arg = arg;

        pdtp__multiplexer_add_socket(mplx, sock, MPLX_READ, nonblock_read_handler, params, 0);
}

static int nonblock_write_handler(pdtp_multiplexer_t mplx, pdtp_sock_t sock, void *arg)
{
        int n, ret;
        nonblock_transfer_parameters_t params = (nonblock_transfer_parameters_t)arg;

        if((n = pdtp__atomic_writesock(sock, params->buffer + params->bytes_transferred, params->bytes_total - params->bytes_transferred)) < 0) {
                pdtp__multiplexer_remove_socket(mplx, sock);
                ret = params->function(mplx, sock, 0, -1, params->arg);

                pdtp__free(params->buffer);
                pdtp__free(params);

                return ret;
        }

        params->bytes_transferred += n;

        /* XXX Add throttling support here */
        if(params->bytes_transferred < params->bytes_total)
                return 1;

        pdtp__multiplexer_remove_socket(mplx, sock);
        ret = params->function(mplx, sock, params->buffer, params->bytes_total, params->arg);

        pdtp__free(params->buffer);
        pdtp__free(params);

        return ret;
}

void pdtp__nonblock_write(pdtp_multiplexer_t mplx, pdtp_sock_t sock, void *buffer, unsigned length, int (*completion_function)(pdtp_multiplexer_t mplx, int sock, void *buffer, int bytes, void *arg), void *arg)
{
        nonblock_transfer_parameters_t params = NEW(nonblock_transfer_parameters);

        params->buffer = pdtp__alloc(length);
        params->bytes_transferred = 0;
        params->bytes_total = length;
        params->function = completion_function;
        params->arg = arg;

        memcpy(params->buffer, buffer, length);

        pdtp__multiplexer_add_socket(mplx, sock, MPLX_WRITE, nonblock_write_handler, params, 0);
}

pdtp_sock_t pdtp__ipv4_tcp_listener_create(uint32_t source_addr, uint16_t port)
{
	pdtp_sock_t sock;
	int val = 1;
	struct sockaddr_in sin;

	memset(&sin, 0, sizeof(struct sockaddr_in));

	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	sin.sin_addr.s_addr = source_addr;

	if((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP)) < 0) 
		return -1;

	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&val, sizeof(val)) < 0) 
		return -1;

	if(bind(sock, (struct sockaddr *)&sin, sizeof(struct sockaddr_in)) < 0) 
		return -1;

	if(listen(sock, MAX_BACKLOG) < 0) 
		return -1;

	return sock;
}

pdtp_sock_t pdtp__ipv4_listener_accept(pdtp_sock_t sock, uint32_t *addr)
{
	pdtp_sock_t connection_sock;
	struct sockaddr_in sin;
	unsigned int len = sizeof(struct sockaddr_in);

	if((connection_sock = accept(sock, (struct sockaddr *)&sin, &len)) < 0)
		return -1;

	if(addr)
		*addr = sin.sin_addr.s_addr;

	return connection_sock;
}
