/* Test driver application for libpdtp 

   Exit Codes:
   0 - All operations completed successfully
   1 - Fatal library error
   2 - Fatal system error

*/

#include "machdep.h"
#include "sockets.h"

#include <stdio.h>
#include <stdlib.h>

#include "alloc.h"
#include "pdtp.h"
#include "socketops.h"
#include "connection.h"

#ifndef WIN32
#include <pthread.h>
#endif

#define LISTEN_ADDR	"127.0.0.1"
#define COMMAND_PORT	2555
#define DATA_PORT	2556

#define PDTP_MAGIC     0x50445450

#ifdef WIN32
typedef HANDLE pdtp_thread_t;
#else
typedef pthread_t *pdtp_thread_t;
#endif /* WIN32 */

pdtp_thread_t thr_create(void *(*thread_func)(void *arg), void *arg)
{
#ifdef WIN32
	return CreateThread(0, 0, thread_func, arg, 0, 0);
#else
	pdtp_thread_t thread = (pdtp_thread_t)pdtp__alloc(sizeof(pthread_t));

	if(pthread_create(thread, 0, thread_func, arg)) {
		free(thread);
		return 0;
	}

	return thread;
#endif /* WIN32 */
}

void thr_exit(void *ptr)
{
#ifdef WIN32
	ExitThread(ptr);
#else
	pthread_exit(ptr);
#endif /* WIN32 */
}

void *testlib_func(void *arg)
{
	pdtp_connection_t conn;
	pdtp_download_t dl;

	printf("--- libpdtp: Connecting to %s:%d\n", LISTEN_ADDR, COMMAND_PORT);

        if(!(conn = pdtp_connect(LISTEN_ADDR, COMMAND_PORT))) {
                puts("failed");
                exit(1);
        } 

	puts("--- libpdtp: Requesting the file '/foo'");

	if(!(dl = pdtp_download_create(conn, "/foo"))) {
		puts("--- libpdtp error: Couldn't retrieve '/foo'");
		exit(1);
	}

	thr_exit(0);

	/* Unreached, but to supress warnings */
	return 0;
}

int cmdfd_handshake(int sock)
{
	uint16_t v16;
	uint32_t v32;

        /* Read the magic number from the client */
        if(pdtp__readsock_int32(sock, &v32) < 0)
		return -1;

	if(v32 != PDTP_MAGIC)
		return -1;

	pdtp__writesock_int32(sock, PDTP_MAGIC);

        /* Minimum protocol version supported, currently hardcoded to 1 */
        pdtp__writesock_int16(sock, 1);

        /* Maximum protocol version supported, currently hardcoded to 1 */
        pdtp__writesock_int16(sock, 1);

        /* Read the client's selected protocol version */
        if(pdtp__readsock_int16(sock, &v16) < 0)
                return -1;

        if(v16 != 1)
                return -1;

	return 0;
}

#if 0
static pdtp_transaction_in_t txn_read(int sock)
{
	pdtp_transaction_in_t txn;
	pdtp_connection_t conn = NEW(pdtp_connection);
	pdtp_notifier_t notifier;

	conn->sock = sock;
	conn->addr = 0;
	conn->socket_lock = pdtp__mutex_create();
	conn->txnmgr = pdtp__transaction_manager_create();
	
	pdtp__mutex_destroy(conn->socket_lock);
	pdtp__transaction_manager_destroy(conn->txnmgr);
}

static int txn_send(int sock, pdtp_transaction_out_t txn)
{
}
#endif

int main()
{
	pdtp_thread_t testlib_thread;
	int listenfd, cmdfd;

	puts("--- | libpdtp test driver starting | ---");

	printf("*** Creating a listener socket on %s:%d... ", LISTEN_ADDR, COMMAND_PORT);
	fflush(stdout);

	if((listenfd = pdtp__ipv4_tcp_listener_create(inet_addr(LISTEN_ADDR), COMMAND_PORT)) < 0) {
		puts("failed.");
		exit(2);
	} else 
		puts("done.");

	puts("*** Running library functions thread");

	if(!(testlib_thread = thr_create(testlib_func, 0))) {
		puts("Error: Couldn't start thread");
		exit(1);
	} else

	if((cmdfd = pdtp__ipv4_listener_accept(listenfd, 0)) < 0) {
		printf("!!! Listener socket error");
		exit(2);
	}

	close(listenfd);
	puts("*** Connection completed.  Doing handshaking...");

	if(cmdfd_handshake(cmdfd) < 0) {
		puts("Error: Handshake failed");
		exit(1);
	}

	puts("*** Handshake succeeded");

	exit(0);
}
