/*
   threads.c - Win32/POSIX thread abstraction layer
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote
 products derived from this software without specific prior written
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "alloc.h"
#include "threadops.h"

#ifdef HAVE_PTHREADS
#include <pthread.h>
#endif

pdtp_mutex_t pdtp__mutex_create(void)
{
#ifdef HAVE_PTHREADS
	pdtp_mutex_t mutex = (pdtp_mutex_t)pdtp__alloc(sizeof(pthread_mutex_t));
	pthread_mutex_init(mutex, 0);
	return mutex;
#endif
	
#ifdef WIN32
	pdtp_mutex_t mutex = (pdtp_mutex_t)pdtp__alloc(sizeof(CRITICAL_SECTION));

	InitializeCriticalSection(mutex);
	
	return mutex;
#endif
	
#if !defined(HAVE_PTHREADS) && !defined(WIN32)
	return 0;
#endif
}

void pdtp__mutex_destroy(pdtp_mutex_t mutex)
{
#ifdef HAVE_PTHREADS
	pthread_mutex_destroy(mutex);
	pdtp__free(mutex);
#endif
#ifdef WIN32
	DeleteCriticalSection(mutex);
	pdtp__free(mutex);
#endif
}

pdtp_cond_t pdtp__cond_create(void)
{
#ifdef HAVE_PTHREADS
	pdtp_cond_t cond = (pdtp_cond_t)pdtp__alloc(sizeof(pthread_cond_t));
	pthread_cond_init(cond, 0);
	return cond;
#endif
#ifdef WIN32
	return CreateEvent(0, TRUE, FALSE, 0);
#endif
#if !defined(HAVE_PTHREADS) && !defined(WIN32)
        return 0;
#endif
}

void pdtp__cond_destroy(pdtp_cond_t cond)
{
#ifdef HAVE_PTHREADS
	pthread_cond_destroy(cond);
	pdtp__free(cond);
#endif
#ifdef WIN32
	CloseHandle(cond);
#endif
}

#ifdef WIN32
int pdtp__cond_wait(pdtp_cond_t cond, pdtp_mutex_t mutex)
{
	pdtp__mutex_unlock(mutex);

	if(WaitForSingleObject(cond, INFINITE) == WAIT_FAILED)
		return -1;

	pdtp__mutex_lock(mutex);

	return 0;
}
#endif /* WIN32 */

#if !defined(HAVE_PTHREADS) && !defined(WIN32)
#warning Building non-thread safe version of libpdtp
int pdtp__mutex_lock(pdtp_mutex_t mutex)
{
	return -1;
}

int pdtp__mutex_unlock(pdtp_mutex_t mutex)
{
	return -1;
}

int pdtp__cond_wait(pdtp_cond_t cond, pdtp_mutex_t mutex)
{
	return -1;
}

int pdtp__cond_signal(pdtp_cond_t cond)
{
	return -1;
}
#endif /* !defined(HAVE_PTHREADS) && !defined(WIN32) */
