/*
   transaction.c - Code for constructing and processing transactions
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.

 * The name of Anthony Arcieri may not be used to endorse or promote
 products derived from this software without specific prior written
 permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "machdep.h"
#include "pdtp.h"
#include "sockets.h"

#include <sys/types.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>

#include "alloc.h"
#include "debug.h"
#include "collection.h"
#include "connection.h"
#include "multiplexer.h"
#include "notifier.h"
#include "socketops.h"
#include "threadops.h"
#include "transaction.h"

/* Include uio.h if we are using UIO */
#ifdef HAVE_UIO
#include <sys/uio.h>
#endif

/* Sanity limit for the size of a transaction (2MB) */
#define	MAX_TXN_SIZE	0x200000

/* Transaction type bit */
#define TXN_REPLY_BIT   0x8000

typedef struct pdtp_reader_list_node {
	pdtp_notifier_t notifier;
	struct pdtp_reader_list_node *next;
} *pdtp_reader_list_node_t;

typedef struct pdtp_object_filter_list_node {
	void *payload;
	unsigned length;
	pdtp_notifier_t notifier;
	struct pdtp_object_filter_list_node *next;
} *pdtp_object_filter_list_node_t;

struct pdtp_transaction_manager {
	/* Connection serial number and respective lock */
	uint32_t serial;
	pdtp_mutex_t serial_lock;

	/* Used for thread safe transaction reading */
	unsigned active_readers;
	pdtp_reader_list_node_t reader_list;
	pdtp_mutex_t reader_lock;

	/* Serial filter collection and respective lock */
	pdtp_collection_t serial_collection;
	pdtp_mutex_t serial_collection_lock;

	/* Object filter collection and respective lock */
	pdtp_collection_t object_collection;
	pdtp_mutex_t object_collection_lock;

	/* Notifier pool and respective lock */
	pdtp_notifier_pool_t notifier_pool;
	pdtp_mutex_t notifier_lock;
};

struct pdtp_transaction_listener {
	pdtp_connection_t conn;
	pdtp_notifier_t notifier;
	pdtp_read_mode_t mode;
	int sock;
};

typedef struct pdtp_transaction_listener_params {
	pdtp_transaction_listener_t listener;
	int (*function)(pdtp_multiplexer_t, pdtp_transaction_listener_t, pdtp_transaction_in_t, void *arg);
	void *arg;
} *pdtp_transaction_listener_params_t;

typedef struct pdtp_transaction_notification {
        pdtp_notifier_t notifier;
        pdtp_transaction_in_t transaction;
} *pdtp_transaction_notification_t;

typedef struct pdtp_object_filter_iterator_data {
	pdtp_notifier_t notifier;
	pdtp_collection_t collection;
} *pdtp_object_filter_iterator_data_t;

struct pdtp_transaction_object {
	uint32_t obj_len;
	uint16_t obj_id;
	void *obj_data;
};

struct pdtp_transaction_in {
	pdtp_transaction_type_t type;
	uint32_t serial;
	uint16_t id;

	uint8_t *buffer;
	uint32_t bytes_read, next_object_offset, length;
	uint16_t current_obj, obj_count, total_objs;

	struct pdtp_transaction_object *object;
	pdtp_connection_t connection;
};

struct pdtp_transaction_out {
	uint32_t length;
	uint32_t serial;
	uint16_t id;

	uint16_t obj_count, max_objects;

#ifdef HAVE_UIO
	struct iovec *iov;
#endif

#ifdef WIN32
	LPWSABUF buf;
#endif

#if !defined(HAVE_UIO) && !defined(WIN32)
	struct pdtp_transaction_object *object;
#endif
};

pdtp_transaction_manager_t pdtp__transaction_manager_create(void)
{
	pdtp_transaction_manager_t txnmgr;

	txnmgr = NEW(pdtp_transaction_manager);

	/* Initialize serial counter and respective lock */
	txnmgr->serial = 1;
	txnmgr->serial_lock = pdtp__mutex_create();

	/* Initialize reader counter and respective lock */
	txnmgr->active_readers = 0;
	txnmgr->reader_lock = pdtp__mutex_create();

	/* Initialize serial filter collection and respective lock */
	txnmgr->serial_collection = pdtp__collection_create();
	txnmgr->serial_collection_lock = pdtp__mutex_create();

	/* Initialize object filter collection and respective lock */
	txnmgr->object_collection = pdtp__collection_create();
	txnmgr->object_collection_lock = pdtp__mutex_create();

	/* Initialize notifier pool and respective lock */
	txnmgr->notifier_pool = pdtp__notifier_pool_create();
	txnmgr->notifier_lock = pdtp__mutex_create();

	return txnmgr;
}

void pdtp__transaction_manager_destroy(pdtp_transaction_manager_t txnmgr)
{
	/* XXX Destroy the object filter collection and associated lock
	   Note a destructor may be needed to prevent memory leaks */
	pdtp__collection_destroy(txnmgr->object_collection, 0);
	pdtp__mutex_destroy(txnmgr->object_collection_lock);

	/* XXX Destroy the serial filter collection and associated lock
	   Note a destructor may be needed to prevent memory leaks */
	pdtp__collection_destroy(txnmgr->serial_collection, 0);
	pdtp__mutex_destroy(txnmgr->serial_collection_lock);

	/* Destroy the notifier pool and associated lock */
	pdtp__notifier_pool_destroy(txnmgr->notifier_pool);
	pdtp__mutex_destroy(txnmgr->notifier_lock);

	/* Destroy remaining locks */
	pdtp__mutex_destroy(txnmgr->reader_lock);
	pdtp__mutex_destroy(txnmgr->serial_lock);

	/* Free the structure itself */
	pdtp__free(txnmgr);
}

/* pdtp__transaction_read_header() establishes an invariant state
   which can then be abstractly utilized in a number of different
   ways in order to complete the process of reading a transaction. */
static pdtp_transaction_in_t pdtp__transaction_read_header(pdtp_connection_t conn) 
{
	pdtp_transaction_in_t txn;
	uint16_t v16;
	uint32_t v32;
	uint8_t header[12];

	/* Read the transaction header */
	if(pdtp__readsock(conn->sock, header, 12) < 0) 
		goto err0;

	txn = NEW(pdtp_transaction_in);

	/* We will use the connection handle later to read more of the
	   transaction and process it as we go */
	txn->connection = conn;

	/* These start at zero and are incremented as we go */
	txn->current_obj = 0;
	txn->obj_count = 0;
	txn->bytes_read = 0;
	txn->next_object_offset = 0;

	/* Extract the transaction length from the header */
	memcpy(&v32, header, 4);
	txn->length = ntohl(v32);

	/* Verify that the transaction does not exceed the maximum length */
	if(txn->length > MAX_TXN_SIZE) 
		goto err1;

	/* Subtract the length of the header itself */
	txn->length -= 8;

	/* Extract the transaction's serial number */
	memcpy(&v32, header + 4, 4);
	txn->serial = ntohl(v32);

	/* Extract the transaction ID */
	memcpy(&v16, header + 8, 2);
	txn->id = ntohs(v16);

	if(txn->id & TXN_REPLY_BIT) {
		txn->type = TXN_REPLY;
		txn->id &= ~TXN_REPLY_BIT;
	} else
		txn->type = TXN_COMMAND;

	/* Extract the object count */
	memcpy(&v16, header + 10, 2);
	txn->total_objs = ntohs(v16);

	if(!txn->total_objs) {
		/* Check for reported garbage at the end of the transaction */
		if(txn->length > 0)
			goto err1;

                /* Nothing left to read.  Unlock the connection */
                pdtp__mutex_unlock(conn->socket_lock);
#ifndef NDEBUG
                debug("--- socket_lock: pdtp__transaction_read_direct()");
#endif

		return txn;
	}

	txn->buffer = (uint8_t *)pdtp__alloc(txn->length);
	txn->object = (struct pdtp_transaction_object *)pdtp__alloc(txn->total_objs * sizeof(struct pdtp_transaction_object));

	return txn;

err1:
	pdtp__free(txn);
err0:
	pdtp__mutex_unlock(conn->socket_lock);

#ifndef NDEBUG
	debug("--- socket_lock: pdtp__transaction_read_header()");
#endif

	return 0;
}

/* Ensure that the transaction's objects are populated to at least the
   specified object index. */
int pdtp__transaction_read_to_object(pdtp_transaction_in_t txn, unsigned obj)
{
	int len;
	uint16_t v16;
	uint32_t v32;

	/* Invalid object index */
	if(obj >= txn->total_objs)
		return -1;

	/* In this case the given object has already been populated */
	if(txn->obj_count >= obj + 1)
		return 0;

	while(txn->obj_count <= obj) {
		/* Transaction is too short for full object header */
		if(txn->length - txn->next_object_offset < 6)
			goto err;

		/* Loop to ensure the buffer contains the object header */
		while(txn->bytes_read - txn->next_object_offset < 6) { 
			if((len = pdtp__timeout_readsock(txn->connection->sock, txn->buffer + txn->bytes_read, txn->length - txn->bytes_read)) < 1)
				goto err;

			txn->bytes_read += len;
		}

		/* Extract the object's length */
		memcpy(&v32, txn->buffer + txn->next_object_offset, 4);
		txn->object[txn->obj_count].obj_len = ntohl(v32);

		txn->next_object_offset += 4;

		/* Extract the object's ID */
		memcpy(&v16, txn->buffer + txn->next_object_offset, 2);
		txn->object[txn->obj_count].obj_id = ntohs(v16);

		txn->next_object_offset += 2;

		/* Transaction is too short to contain the full object */
		if(txn->length - txn->next_object_offset < txn->object[txn->obj_count].obj_len)
			goto err;

		/* Loop to ensure the buffer contains the object payload */
		while(txn->bytes_read - txn->next_object_offset < txn->object[txn->obj_count].obj_len) { 
			if((len = pdtp__timeout_readsock(txn->connection->sock, txn->buffer + txn->bytes_read, txn->length - txn->bytes_read)) < 1) 
				goto err;

			txn->bytes_read += len;
		}

		txn->object[txn->obj_count].obj_data = txn->buffer + txn->next_object_offset;
		txn->next_object_offset += txn->object[txn->obj_count].obj_len;
		txn->obj_count++;
	}

	/* After the entire transaction has been read some additional protocol
	   validation can be performed, after which the connection is
	   unlocked so the next transaction can be read */
	if(txn->obj_count == txn->total_objs) {
		/* Garbage on the end of the transaction is a protocol
		   violation */
		if(txn->bytes_read != txn->length)
			goto err;

		pdtp__mutex_unlock(txn->connection->socket_lock);
#ifndef NDEBUG
		debug("--- socket_lock: pdtp__transaction_read_to_object()");
#endif
		txn->connection = 0;
	}

	return 0;

err:
	/* XXX At this point we've lost the connection or encountered a
	   protocol error at which point we should mark the connection 
	   as lost */
	
	pdtp__mutex_unlock(txn->connection->socket_lock);
#ifndef NDEBUG
	debug("--- socket_lock: pdtp__transaction_read_to_object()");
#endif
	
	txn->connection = 0;

	return -1;
}

/* This function is invoked by pdtp__transaction_set_serial automatically and
   thus does not need to be exposed outside of this module */
static pdtp_notifier_t pdtp__create_serial_notifier(pdtp_connection_t conn, uint32_t serial)
{
	pdtp_notifier_t notifier;

	pdtp__mutex_lock(conn->txnmgr->notifier_lock);
	notifier = pdtp__notifier_get(conn->txnmgr->notifier_pool);
	pdtp__mutex_unlock(conn->txnmgr->notifier_lock);

	pdtp__mutex_lock(conn->txnmgr->serial_collection_lock);
	pdtp__collection_insert(conn->txnmgr->serial_collection, serial, notifier);
	pdtp__mutex_unlock(conn->txnmgr->serial_collection_lock);

	return notifier;
}

pdtp_notifier_t pdtp__create_object_notifier(pdtp_connection_t conn, uint16_t id, void *payload, unsigned length)
{
	pdtp_notifier_t notifier;
	pdtp_object_filter_list_node_t t, n;

#ifndef NDEBUG
	debug("Creating object notifier for %d", id);
#endif

	pdtp__mutex_lock(conn->txnmgr->notifier_lock);
	notifier = pdtp__notifier_get(conn->txnmgr->notifier_pool);
	pdtp__mutex_unlock(conn->txnmgr->notifier_lock);

	n = NEW(pdtp_object_filter_list_node);
	n->next = 0;
	n->length = length;
	n->notifier = notifier;
	n->payload = pdtp__alloc(length);
	memcpy(n->payload, payload, length);

	pdtp__mutex_lock(conn->txnmgr->object_collection_lock);
	if((t = (pdtp_object_filter_list_node_t)pdtp__collection_remove(conn->txnmgr->object_collection, id)))
		n->next = t;

	pdtp__collection_insert(conn->txnmgr->object_collection, id, n);
	pdtp__mutex_unlock(conn->txnmgr->object_collection_lock);

	return notifier;
}

/* XXX Unit tests should be performed on the collection code to ensure that
   removing an entry from a collection within an iterator works correctly */
static int pdtp__destroy_object_notifier_filter(pdtp_collection_key_t key, void *data, void *ptr)
{
	pdtp_object_filter_iterator_data_t params = (pdtp_object_filter_iterator_data_t)ptr;
	pdtp_object_filter_list_node_t temp, node, root = (pdtp_object_filter_list_node_t)data;

	if(root->notifier == params->notifier) {
		node = root;
		root = root->next;
		
		if(!root)
			pdtp__collection_remove(params->collection, key);
	} else {
		for(node = root; node->next && node->next->notifier != params->notifier; node = node->next);

		/* If the notifier was not found continue looking */
		if(!node->next)
			return 1;

		temp = node;
		node = node->next;
		temp->next = node->next;
	}

	/* Node is now set to the filter rule we wish to remove; destroy it */
	pdtp__free(node->payload);
	pdtp__free(node);

	return 0;
}

void pdtp__destroy_object_notifier(pdtp_connection_t conn, pdtp_notifier_t notifier)
{
	struct pdtp_object_filter_iterator_data data;

	/* Remove the associated object filter rules */
	pdtp__mutex_lock(conn->txnmgr->object_collection_lock);
	data.notifier = notifier;
	data.collection = conn->txnmgr->object_collection;
	
	pdtp__collection_iterate(conn->txnmgr->object_collection, pdtp__destroy_object_notifier_filter, &data);
	pdtp__mutex_unlock(conn->txnmgr->object_collection_lock);

	/* Return the notifier to the pool */
	pdtp__mutex_lock(conn->txnmgr->notifier_lock);
	pdtp__notifier_return(conn->txnmgr->notifier_pool, notifier);
	pdtp__mutex_unlock(conn->txnmgr->notifier_lock);
}

/* This function scans the current set of filter rules for the given connection
   and returns the first notifier which matches one of the rules */
static pdtp_notifier_t pdtp__transaction_filter(pdtp_connection_t conn, pdtp_transaction_in_t txn)
{
	unsigned i;
	void *data;
	uint16_t id;
	uint32_t length;

	pdtp_object_filter_list_node_t node;
	pdtp_notifier_t notifier = 0;

	if(txn->serial) {
		pdtp__mutex_lock(conn->txnmgr->serial_collection_lock);
		notifier = (pdtp_notifier_t)pdtp__collection_remove(conn->txnmgr->serial_collection, txn->serial);
		pdtp__mutex_unlock(conn->txnmgr->serial_collection_lock);

		return notifier;
	} 
	pdtp__mutex_lock(conn->txnmgr->object_collection_lock);

	/* Scan all objects in the transaction looking for a matching
	   object filter rule */
	for(i = 0; i < pdtp__transaction_object_count(txn); i++) {
		if(pdtp__transaction_object_id(txn, i, &id) < 0)
			goto err;

#ifndef NDEBUG
                debug("Scanning for object filter on %d", id);
#endif

		if(!(node = (pdtp_object_filter_list_node_t)pdtp__collection_lookup(conn->txnmgr->object_collection, id)))
			continue;

#ifndef NDEBUG
		debug("Found matching object filter chain!");
#endif

		if(pdtp__transaction_object_data(txn, i, &data) < 0)
			goto err;

		if(pdtp__transaction_object_length(txn, i, &length) < 0)
			goto err;

		while(node) {
#ifndef NDEBUG
			debug("Comparing %d bytes to node with length %d bytes", length, node->length);
#endif
			if(node->length == length && !memcmp(data, node->payload, length))
				break;

			node = node->next;
		}

#ifndef NDEBUG
		if(node)
			debug("Found matching object filter rule!");
		else
			debug("No matching object filter rule");
#endif
		if(node) {
			notifier = node->notifier;
			break;
		}
	}

	pdtp__mutex_unlock(conn->txnmgr->object_collection_lock);
	return notifier;

err:
	pdtp__mutex_unlock(conn->txnmgr->object_collection_lock);
	return 0;
}

static pdtp_transaction_in_t pdtp__transaction_read_direct(pdtp_connection_t conn, pdtp_notifier_t notifier, pdtp_read_mode_t mode)
{
	pdtp_transaction_in_t txn = 0;
	pdtp_notifier_t dest_notifier = 0;
	pdtp_transaction_notification_t msg;
	pdtp_reader_list_node_t node;

	/* The lock on the connection may be held by another thread which
	   is in the process of reading a transaction in PDTP_FETCH mode.
	   We lock the connection here to ensure the transaction is read
	   in full first before continuing */
	while(notifier != dest_notifier) {
#ifndef NDEBUG
		debug("^^^ socket_lock: pdtp__transaction_read_direct()");
#endif
		pdtp__mutex_lock(conn->socket_lock);
#ifndef NDEBUG
		debug("+++ socket_lock: pdtp__transaction_read_direct()");
#endif

		if(!(txn = pdtp__transaction_read_header(conn)))
			goto err;

		if(mode == PDTP_STORE && txn->total_objs) {
			if(pdtp__transaction_read_to_object(txn, txn->total_objs - 1) < 0)
				goto err;
		} 

		/* If the transaction matched no filters discard it */
		if(!(dest_notifier = pdtp__transaction_filter(conn, txn)))
			pdtp__transaction_in_destroy(txn);
		else if(notifier != dest_notifier) {
			msg = NEW(pdtp_transaction_notification);
			msg->notifier = notifier;
			msg->transaction = txn;

			pdtp__notifier_send_message(dest_notifier, msg);

			/* Here we release the connection lock if we were 
			   holding it and wait for the other thread to 
			   aacquire it */
			if(txn->connection) {
				pdtp__mutex_unlock(conn->socket_lock);
#ifndef NDEBUG
				debug("--- socket_lock: pdtp__transaction_read_direct()");
#endif
				
				pdtp__notifier_recv_message(notifier);
			}
		}
	}

	/* We have now acquired the transaction we were waiting for.  We can
	   safely return the notifier to the pool if we were waiting on
	   a synchronous transaction */
	if(txn->serial) {
		pdtp__mutex_lock(conn->txnmgr->notifier_lock);
		pdtp__notifier_return(conn->txnmgr->notifier_pool, notifier);
		pdtp__mutex_unlock(conn->txnmgr->notifier_lock);
	}

	/* Decrementing of the readers counter and selection of a new thread
	   to receive the transaction directly are handled in this function */
	pdtp__mutex_lock(conn->txnmgr->reader_lock);
	if(--conn->txnmgr->active_readers) {
		/* Select a new thread to handle receiving the transaction */
		node = conn->txnmgr->reader_list;
		conn->txnmgr->reader_list = node->next;

		pdtp__notifier_send_message(node->notifier, 0);
		pdtp__free(node);
	}
	pdtp__mutex_unlock(conn->txnmgr->reader_lock);

	return txn;	
err:
	/* XXX Mark the connection as bad and return 0 */
	return 0;
}

pdtp_transaction_in_t pdtp__transaction_wait(pdtp_connection_t conn, pdtp_notifier_t notifier, pdtp_read_mode_t mode)
{
	pdtp_reader_list_node_t node, tempnode;
	pdtp_transaction_in_t txn;
	pdtp_transaction_notification_t msg;

	pdtp__mutex_lock(conn->txnmgr->reader_lock);

	if(!conn->txnmgr->active_readers++) {
		pdtp__mutex_unlock(conn->txnmgr->reader_lock);

		return pdtp__transaction_read_direct(conn, notifier, mode);
	}

	node = NEW(pdtp_reader_list_node);
	node->next = conn->txnmgr->reader_list;
	conn->txnmgr->reader_list = node->next;

	node->notifier = notifier;

	pdtp__mutex_unlock(conn->txnmgr->reader_lock);

	/* A NULL message indicates this thread has been selected to read the
	   incoming transaction directly */
	if(!(msg = (pdtp_transaction_notification_t)pdtp__notifier_recv_message(notifier))) 
		return pdtp__transaction_read_direct(conn, notifier, mode);

	txn = msg->transaction;

	/* If the transaction is being received in PDTP_FETCH mode we set this 
	   thread to own the lock on the connection socket and notify the
	   other thread that we are now holding the lock */
	if(txn->connection) {
		pdtp__mutex_lock(conn->socket_lock);
#ifndef NDEBUG
		debug("--- socket_lock: pdtp__transaction_wait()");
#endif
		pdtp__notifier_send_message(msg->notifier, 0);
	}

	/* Free the message datagram */
	pdtp__free(msg);

	pdtp__mutex_lock(conn->txnmgr->reader_lock);

	conn->txnmgr->active_readers--;

	/* Remove us from the list of active readers */
	if(notifier == conn->txnmgr->reader_list->notifier) {
		node = conn->txnmgr->reader_list;
		conn->txnmgr->reader_list = node->next;

		pdtp__free(node);
	} else {
		for(node = conn->txnmgr->reader_list; node->next && node->next->notifier != notifier; node = node->next);

		tempnode = node->next;
		node->next = tempnode->next;

		pdtp__free(tempnode);
	}

	pdtp__mutex_unlock(conn->txnmgr->reader_lock);

	return txn;
}

pdtp_transaction_listener_t pdtp__transaction_listener_create(pdtp_connection_t conn, pdtp_notifier_t notifier, pdtp_read_mode_t mode)
{
	pdtp_reader_list_node_t node;
	pdtp_transaction_listener_t listener = NEW(pdtp_transaction_listener);

	listener->conn = conn;
	listener->notifier = notifier;
	listener->mode = mode;

	pdtp__mutex_lock(conn->txnmgr->reader_lock);
	if(!conn->txnmgr->active_readers++) {
		listener->sock = conn->sock;
		pdtp__mutex_unlock(conn->txnmgr->reader_lock);
		return listener;
	}

	listener->sock = -1;

	node = NEW(pdtp_reader_list_node);
	node->next = conn->txnmgr->reader_list;
	conn->txnmgr->reader_list = node->next;

	node->notifier = notifier;

	pdtp__mutex_unlock(conn->txnmgr->reader_lock);

	return listener;
}

void pdtp__transaction_listener_destroy(pdtp_transaction_listener_t listener)
{
	pdtp_reader_list_node_t node, tempnode;

	/* XXX Handle multiplexer removal */

	if(listener->sock != -1) {
		pdtp__mutex_lock(listener->conn->txnmgr->reader_lock);
		if(--listener->conn->txnmgr->active_readers) {
			/* Select a new thread to handle receiving the 
			   transaction */
			node = listener->conn->txnmgr->reader_list;
			listener->conn->txnmgr->reader_list = node->next;

			pdtp__notifier_send_message(node->notifier, 0);
			pdtp__free(node);
		}
		pdtp__mutex_unlock(listener->conn->txnmgr->reader_lock);

		pdtp__free(listener);
		return;
	}

	pdtp__mutex_lock(listener->conn->txnmgr->reader_lock);
	listener->conn->txnmgr->active_readers--;

	/* Remove notifier from the list of active readers */
	if(listener->notifier == listener->conn->txnmgr->reader_list->notifier) {
		node = listener->conn->txnmgr->reader_list;
		listener->conn->txnmgr->reader_list = node->next;

		pdtp__free(node);
	} else {
		for(node = listener->conn->txnmgr->reader_list; node->next && node->next->notifier != listener->notifier; node = node->next);

		tempnode = node->next;
		node->next = tempnode->next;

		pdtp__free(tempnode);
	}

	pdtp__mutex_unlock(listener->conn->txnmgr->reader_lock);

	pdtp__free(listener);
}

static void pdtp__serial_listener_cleanup(pdtp_multiplexer_t mplx, pdtp_transaction_listener_t listener)
{
        pdtp_connection_t conn = listener->conn;
        pdtp_notifier_t notifier = listener->notifier;

        pdtp__multiplexer_remove_transaction_listener(mplx, listener);
        pdtp__transaction_listener_destroy(listener);

        pdtp__mutex_lock(conn->txnmgr->notifier_lock);
        pdtp__notifier_return(conn->txnmgr->notifier_pool, notifier);
        pdtp__mutex_unlock(conn->txnmgr->notifier_lock);
}

static int pdtp__transaction_socket_handler(pdtp_multiplexer_t mplx, int sock, void *arg)
{
	int ret, cleanup = 0;
	
	pdtp_transaction_listener_params_t params = (pdtp_transaction_listener_params_t)arg;
	pdtp_transaction_in_t txn;
	pdtp_notifier_t dest_notifier = 0;
	pdtp_transaction_notification_t msg;

#ifndef NDEBUG
	debug("^^^ socket_lock: pdtp__transaction_socket_handler()");
#endif

	/* The lock on the connection may be held by another thread which
	   is in the process of reading a transaction in PDTP_FETCH mode.
	   We lock the connection here to ensure the transaction is read
	   in full first before continuing */
	pdtp__mutex_lock(params->listener->conn->socket_lock);

#ifndef NDEBUG
	debug("+++ socket_lock: pdtp__transaction_socket_handler()");
	debug("Reading transaction header...");
#endif

	if(!(txn = pdtp__transaction_read_header(params->listener->conn)))
		goto err;

#ifndef NDEBUG
	debug("Reading transaction objects...");
#endif

	if(params->listener->mode == PDTP_STORE && txn->total_objs) {
		if(pdtp__transaction_read_to_object(txn, txn->total_objs - 1) < 0)
			goto err;
	}

#ifndef NDEBUG
	debug("Looking up transaction in filter...");
#endif

	/* If the incoming transaction matched no filters discard it */
	if(!(dest_notifier = pdtp__transaction_filter(params->listener->conn, txn))) {
#ifndef NDEBUG
		debug("!!! Transaction doesn't match any registered filters");
#endif
		pdtp__transaction_in_destroy(txn);
		return 1;
	}

	/* If the transaction was not intended for us, pass it along */
	if(params->listener->notifier != dest_notifier) {
#ifndef NDEBUG
		debug("Filter matches foreign thread");
#endif
		msg = NEW(pdtp_transaction_notification);
		msg->notifier = params->listener->notifier;
		msg->transaction = txn;

		pdtp__notifier_send_message(dest_notifier, msg);

		/* Here we release the connection lock if we were 
		   holding it and wait for the other thread to 
		   aacquire it */
		if(txn->connection) {
			pdtp__mutex_unlock(params->listener->conn->socket_lock);
#ifndef NDEBUG
			debug("--- socket_lock: pdtp__transaction_socket_handler()");
#endif

			pdtp__notifier_recv_message(params->listener->notifier);
		}

		return 1;
	}

#ifndef NDEBUG
	debug("Filter matches our thread");
#endif

	/* If we are waiting for a serial schedule cleanup */
	if(txn->serial)
		cleanup = 1;
	
	/* Otherwise call the incoming transaction handler */
	ret = params->function(mplx, params->listener, txn, params->arg);

	if(cleanup)
		pdtp__serial_listener_cleanup(mplx, params->listener);

	return ret;
err:
	return 0;
}

static int pdtp__transaction_notifier_handler(pdtp_multiplexer_t mplx, pdtp_notifier_t notifier, void *arg)
{
	int ret, cleanup = 0;

	pdtp_transaction_listener_params_t params = (pdtp_transaction_listener_params_t)arg;
        pdtp_transaction_in_t txn;
	pdtp_transaction_notification_t msg;

        /* A NULL message indicates this thread has been selected to read the
           incoming transaction directly */
        if(!(msg = (pdtp_transaction_notification_t)pdtp__notifier_recv_message(notifier))) {
		params->listener->sock = params->listener->conn->sock;
		
		pdtp__multiplexer_remove_notifier(mplx, notifier);
		pdtp__multiplexer_add_socket(mplx, params->listener->sock, MPLX_READ, pdtp__transaction_socket_handler, params, 0);

		return 1;
	}

        txn = msg->transaction;

        /* If the transaction is being received in PDTP_FETCH mode we set this 
           thread to own the lock on the connection socket and notify the
           other thread that we are now holding the lock */
        if(txn->connection) {
#ifndef NDEBUG
		debug("^^^ socket_lock: pdtp__transaction_notifier_handler()");
#endif
		
                pdtp__mutex_lock(params->listener->conn->socket_lock);

#ifndef NDEBUG
		debug("+++ socket_lock: pdtp__transaction_notifier_handler()");
#endif

                pdtp__notifier_send_message(msg->notifier, 0);
        }

        /* Free the message datagram */
        pdtp__free(msg);

        /* If we are waiting for a serial schedule cleanup */
        if(txn->serial)
                cleanup = 1;

	ret = params->function(mplx, params->listener, txn, params->arg);

	if(cleanup)
		pdtp__serial_listener_cleanup(mplx, params->listener);

	return ret;
}

void pdtp__multiplexer_add_transaction_listener(pdtp_multiplexer_t mplx, pdtp_transaction_listener_t listener, int (*function)(pdtp_multiplexer_t, pdtp_transaction_listener_t, pdtp_transaction_in_t, void *arg), void *arg)
{
	pdtp_transaction_listener_params_t params = NEW(pdtp_transaction_listener_params);

	params->listener = listener;
	params->function = function;
	params->arg = arg;

	if(listener->sock != -1) {
		pdtp__multiplexer_add_socket(mplx, listener->sock, MPLX_READ, pdtp__transaction_socket_handler, params, 0);
	} else
		pdtp__multiplexer_add_notifier(mplx, listener->notifier, pdtp__transaction_notifier_handler, params);
}

void *pdtp__multiplexer_remove_transaction_listener(pdtp_multiplexer_t mplx, pdtp_transaction_listener_t listener)
{
	void *ret;
	pdtp_transaction_listener_params_t params;

	if(listener->sock != -1)
		params = (pdtp_transaction_listener_params_t)pdtp__multiplexer_remove_socket(mplx, listener->sock);
	else {
		params = (pdtp_transaction_listener_params_t)pdtp__multiplexer_remove_notifier(mplx, listener->notifier);
	}

	ret = params->arg;
	pdtp__free(params);

	return ret;
}

void pdtp__transaction_in_destroy(pdtp_transaction_in_t txn)
{
	if(txn->total_objs) {
		if(txn->connection)
			pdtp__transaction_read_to_object(txn, txn->total_objs - 1);

		pdtp__free(txn->object); 
		pdtp__free(txn->buffer);
	}

	pdtp__free(txn);
}

pdtp_transaction_type_t pdtp__transaction_type(pdtp_transaction_in_t txn)
{
	return txn->type;
}

uint16_t pdtp__transaction_id(pdtp_transaction_in_t txn)
{
	return txn->id;
}

uint32_t pdtp__transaction_serial(pdtp_transaction_in_t txn)
{
	return txn->serial;
}

unsigned pdtp__transaction_object_count(pdtp_transaction_in_t txn)
{
	return txn->obj_count;
}

int pdtp__transaction_object_index(pdtp_transaction_in_t txn, uint16_t id)
{
	while(txn->current_obj < txn->total_objs) { 
		if(pdtp__transaction_read_to_object(txn, txn->current_obj) < 0) 
			return -1;

		if(txn->object[txn->current_obj].obj_id >= id)
			break;

		txn->current_obj++;
	}

	if(txn->current_obj == txn->total_objs || txn->object[txn->current_obj].obj_id != id)
		return -1;
	
#ifndef NDEBUG
	debug("txn->object[%d] is %d", txn->current_obj, id);
#endif
	
	return txn->current_obj;
}

int pdtp__transaction_object_id(pdtp_transaction_in_t txn, unsigned objno, uint16_t *id)
{
	if(pdtp__transaction_read_to_object(txn, objno) < 0)
		return -1;

	*id = txn->object[objno].obj_id;

	return 0;
}

int pdtp__transaction_object_length(pdtp_transaction_in_t txn, unsigned objno, uint32_t *length)
{
	if(pdtp__transaction_read_to_object(txn, objno) < 0)
		return -1;

	*length = txn->object[objno].obj_len;

	return 0;
}

int pdtp__transaction_object_data(pdtp_transaction_in_t txn, unsigned objno, void **data)
{
	if(pdtp__transaction_read_to_object(txn, objno) < 0)
		return -1;

	*data = txn->object[objno].obj_data;

	return 0;
}

int pdtp__transaction_object_uint16(pdtp_transaction_in_t txn, unsigned objno, uint16_t *object)
{
	uint16_t value;
	uint32_t length;
	void *data;

	if(pdtp__transaction_object_length(txn, objno, &length) < 0)
		return -1;

	if(length != 2)
		return -1;

	if(pdtp__transaction_object_data(txn, objno, &data) < 0)
		return -1;

	memcpy(&value, data, 2);
	*object = ntohs(value);

	return 0;
}

int pdtp__transaction_object_uint32(pdtp_transaction_in_t txn, unsigned objno, uint32_t *object)
{
	uint32_t value, length;
	void *data;

	if(pdtp__transaction_object_length(txn, objno, &length) < 0)
		return -1;

	if(length != 4)
		return -1;

	if(pdtp__transaction_object_data(txn, objno, &data) < 0)
		return -1;

	memcpy(&value, data, 4);
	*object = ntohl(value);

	return 0;
}

int pdtp__transaction_object_uint64(pdtp_transaction_in_t txn, unsigned objno, uint64_t *object)
{
	uint64_t value;
	uint32_t length;
	uint8_t *data;

#ifndef HOST_BIGENDIAN
	int i;
#endif

	if(pdtp__transaction_object_length(txn, objno, &length) < 0)
		return -1;

	if(length != 8)
		return -1;

	if(pdtp__transaction_object_data(txn, objno, (void **)&data) < 0)
		return -1;

#ifdef HOST_BIGENDIAN
	memcpy(&value, data, 8);
#else
	for(i = 0; i < 8; i++)
		((uint8_t *)&value)[i] = data[7 - i];
#endif /* HOST_BIGENDIAN */

	*object = value;

	return 0;
}

int pdtp__transaction_object_string(pdtp_transaction_in_t txn, unsigned objno, char **string)
{
	uint32_t length;
	void *data;

	
	if(pdtp__transaction_object_length(txn, objno, &length) < 0)
		return -1;

	if(pdtp__transaction_object_data(txn, objno, &data) < 0)
		return -1;

	*string = (char *)pdtp__alloc(length + 1);
	memcpy(*string, data, length);
	(*string)[length] = '\0';

	return 0;
}

int pdtp__transaction_object_unpack(pdtp_transaction_in_t txn, unsigned objno, const char *format, ...)
{
#ifndef HOST_BIGENDIAN
	int i;
#endif
	int ret = 0;
	uint32_t length;
	uint8_t *bptr;
	uint16_t v16;
	uint32_t v32;
	uint64_t v64;
	va_list ap;
	char *str;

	if(pdtp__transaction_object_length(txn, objno, &length) < 0)
		return -1;

	if(pdtp__transaction_object_data(txn, objno, (void **)&bptr) < 0)
		return -1;

	va_start(ap, format);

	while(*format) {
		switch(*format) {
			case '1':
				if(length < 1)
					goto done;

				*va_arg(ap, int *) = *bptr;

				length -= 1;
				bptr += 1;

				break;
			case '2':
				if(length < 2)
					goto done;

				memcpy(&v16, bptr, 2);
				*va_arg(ap, int *) = htons(v16);

				length -= 2;
				bptr += 2;

				break;
			case '4':
				if(length < 4)
					goto done;

				memcpy(&v32, bptr, 4);
				*va_arg(ap, uint32_t *) = htonl(v32);

				length -= 4;
				bptr += 4;

				break;
			case '8':
				if(length < 8)
					goto done;

#ifdef HOST_BIGENDIAN
				memcpy(&v64, bptr, 8);
#else
				for(i = 0; i < 8; i++)
					((uint8_t *)&v64)[i] = bptr[7 - i];
#endif
				*va_arg(ap, uint64_t *) = v64;

				length -= 8;
				bptr += 8;

				break;
			case 's':
				if(length < 2)
					goto done;

				/* Extract the string length into v16 */
				memcpy(&v16, bptr, 2);
				v16 = htons(v16);

				length -= 2;
				bptr += 2;

				if(length < v16)
					goto done;

				str = (char *)pdtp__alloc(v16 + 1);
				memcpy(str, bptr, v16);
				str[v16] = '\0';
				*va_arg(ap, char **) = str;

				length -= v16;
				bptr += v16;
		}

		format++;
		ret++;
	}

done:
	va_end(ap);
	return ret;
}

pdtp_transaction_out_t pdtp__transaction_create(uint16_t tid, uint16_t obj_count)
{
	pdtp_transaction_out_t txn = NEW(pdtp_transaction_out);

	txn->serial = 0;
	txn->id = htons(tid);

	txn->length = 8;
	txn->obj_count = 0;
	txn->max_objects = obj_count;

#ifdef HAVE_UIO
	txn->iov = (struct iovec *)pdtp__alloc((4 + txn->max_objects) * sizeof(struct iovec));

	/* This iov_base field gets filled in before the writev call */
	txn->iov[0].iov_len = 4;

	txn->iov[1].iov_len = 4;
	txn->iov[1].iov_base = (char *)&txn->serial;

	txn->iov[2].iov_len = 2;
	txn->iov[2].iov_base = (char *)&txn->id;

	/* This iov_base field gets filled in before the writev call */
	txn->iov[3].iov_len = 2;
#endif 

#ifdef WIN32
	txn->buf = (LPWSABUF)pdtp__alloc((4 + txn->max_objects) * sizeof(WSABUF));

	/* This buf field gets filled in before the WSASend call */
	txn->buf[0].len = 4;

	txn->buf[1].len = 4;
	txn->buf[1].buf = (char *)&txn->serial;

	txn->buf[2].len = 2;
	txn->buf[2].buf = (char *)&txn->id;

	/* This buf field gets filled in before the WSASend call */
	txn->buf[3].len = 2;
#endif

#if !defined(HAVE_UIO) && !defined(WIN32)
	txn->object = (struct pdtp_transaction_object *)pdtp__alloc(txn->max_objects * sizeof(struct pdtp_transaction_object));
#endif 

	return txn;
}

void pdtp__transaction_out_destroy(pdtp_transaction_out_t txn)
{
	int i;

#ifdef HAVE_UIO
	for(i = 0; i < txn->obj_count; i++)
		pdtp__free(txn->iov[4 + i].iov_base);

	pdtp__free(txn->iov);
#endif 

#ifdef WIN32
	for(i = 0; i < txn->obj_count; i++)
		pdtp__free(txn->buf[4 + i].buf);

	pdtp__free(txn->buf);
#endif

#if !defined(HAVE_UIO) && !defined(WIN32)
	for(i = 0; i < txn->obj_count; i++)
		pdtp__free(txn->object[i].obj_data);

	pdtp__free(txn->object);
#endif

	pdtp__free(txn);
}

pdtp_notifier_t pdtp__transaction_set_serial(pdtp_transaction_out_t txn, pdtp_connection_t conn)
{
	uint32_t serial;

	pdtp__mutex_lock(conn->txnmgr->serial_lock);
	serial = conn->txnmgr->serial++;
	txn->serial = htonl(serial);
	pdtp__mutex_unlock(conn->txnmgr->serial_lock);

	return pdtp__create_serial_notifier(conn, serial);
}

void pdtp__transaction_add_object(pdtp_transaction_out_t txn, uint16_t id, const void *data, uint32_t length)
{
	uint16_t v16;
	uint32_t v32;
	uint8_t *obj_buf;

	if(txn->obj_count >= txn->max_objects) 
		return;

#if defined(HAVE_UIO) || defined(WIN32)
	obj_buf = (uint8_t *)pdtp__alloc(6 + length);

	v32 = htonl(length);
	memcpy(obj_buf, &v32, 4);
	v16 = htons(id);
	memcpy(obj_buf + 4, &v16, 2);

	if(length > 0) 
		memcpy(obj_buf + 6, data, length);

#ifdef HAVE_UIO
	txn->iov[txn->obj_count + 4].iov_len = 6 + length;
	txn->iov[txn->obj_count + 4].iov_base = (char *)obj_buf;
#endif

#ifdef WIN32
	txn->buf[txn->obj_count + 4].len = 6 + length;
	txn->buf[txn->obj_count + 4].buf = (char *)obj_buf;
#endif

#else
	txn->object[txn->obj_count].obj_len = length;
	txn->object[txn->obj_count].obj_id = id;
	txn->object[txn->obj_count].obj_data = pdtp__alloc(length);
	memcpy(txn->object[txn->obj_count].obj_data, data, length);
#endif 

	txn->obj_count++;
	txn->length += 6 + length;
}

void pdtp__transaction_add_int16(pdtp_transaction_out_t txn, uint16_t id, uint16_t value)
{
	uint16_t v = htons(value);
	pdtp__transaction_add_object(txn, id, &v, 2);
}

void pdtp__transaction_add_int32(pdtp_transaction_out_t txn, uint16_t id, uint32_t value)
{
	uint32_t v = htonl(value);
	pdtp__transaction_add_object(txn, id, &v, 4);
}

void pdtp__transaction_add_string(pdtp_transaction_out_t txn, uint16_t id, const char *string)
{
	pdtp__transaction_add_object(txn, id, string, (uint16_t)strlen(string));
}

int pdtp__transaction_object_pack(pdtp_transaction_out_t txn, uint16_t id, const char *format, ...)
{
	typedef struct pack_node {
		unsigned length;
		uint8_t *buf;

		struct pack_node *link;
	} *pack_node_t;

	pack_node_t head = 0, tail = 0, node;

	unsigned l, length = 0;
	void *buf;
	uint8_t *bptr;
	uint16_t v16;
	uint32_t v32;
	uint64_t v64;
	va_list ap;

#ifndef HOST_BIGENDIAN
	int i;
#endif

	va_start(ap, format);

	while(*format) {
		node = NEW(pack_node);
		node->link = 0;

		switch(*format) {
			case '1':
				length += 1;
				node->length = 1;
				node->buf = (uint8_t *)pdtp__alloc(1);
				*node->buf = va_arg(ap, int);
				break;
			case '2':
				length += 2;
				node->length = 2;
				node->buf = (uint8_t *)pdtp__alloc(2);
				v16 = htons((uint16_t)va_arg(ap, int));
				memcpy(node->buf, &v16, 2);
				break;
			case '4':
				length += 4;
				node->length = 4;
				node->buf = (uint8_t *)pdtp__alloc(4);
				v32 = htonl(va_arg(ap, uint32_t));
				memcpy(node->buf, &v32, 4);
				break;
			case '8':
				length += 8;
				node->length = 8;
				node->buf = (uint8_t *)pdtp__alloc(8);
				v64 = va_arg(ap, uint64_t);
#ifdef HOST_BIGENDIAN
				memcpy(node->buf, &v64, 8);
#else
				for(i = 0; i < 8; i++)
					node->buf[7 - i] = ((uint8_t *)&v64)[i];
#endif
				break;
			case 's':
				if(!(bptr = (uint8_t *)va_arg(ap, char *)))
					bptr = (uint8_t *)"";

				l = (unsigned)strlen((char *)bptr);
				length += l + 2;
				node->length = l + 2;
				node->buf = (uint8_t *)pdtp__alloc(l + 2);
				v16 = htons((uint16_t)l);
				memcpy(node->buf, &v16, 2);
				memcpy(node->buf + 2, bptr, l);
				break;
		}

		if(!head)
			head = tail = node;
		else {
			tail->link = node;
			tail = tail->link;
		}

		format++;
	}

	bptr = buf = pdtp__alloc(length);

	while((node = head)) {
		head = node->link;

		memcpy(bptr, node->buf, node->length);
		bptr += node->length;

		pdtp__free(node->buf);
		pdtp__free(node);
	}

	pdtp__transaction_add_object(txn, id, buf, length);
	pdtp__free(buf);
	va_end(ap);

	return length;
}

int pdtp__transaction_write(pdtp_connection_t conn, pdtp_transaction_out_t txn)
{
#ifdef HAVE_UIO
	uint32_t length;
	uint16_t obj_count;

	length = htonl(txn->length);
	obj_count = htons(txn->obj_count);

	txn->iov[0].iov_base = (char *)&length;
	txn->iov[3].iov_base = (char *)&obj_count;

	return writev(conn->sock, txn->iov, 4 + txn->obj_count);
#endif

#ifdef WIN32
	uint32_t length;
	uint16_t obj_count;
	DWORD bytes = 0;

	length = htonl(txn->length);
	obj_count = htons(txn->obj_count);

	txn->buf[0].buf = (char *)&length;
	txn->buf[3].buf = (char *)&obj_count;

	if(WSASend(conn->sock, txn->buf, 4 + txn->obj_count, &bytes, 0, 0, 0) < 0) 
		return -1;

	return bytes;
#endif

#if !defined(HAVE_UIO) && !defined(WIN32)
	int i, ret;
	uint8_t *buffer, *bptr;
	uint16_t v16;
	uint32_t v32;

	buffer = (uint8_t *)pdtp__alloc(txn->length + 4);

	v32 = htonl(txn->length);
	memcpy(buffer, &v32, 4);

	memcpy(buffer + 4, &txn->serial, 4);
	memcpy(buffer + 8, &txn->id, 2);

	v16 = htons(txn->obj_count);
	memcpy(buffer + 10, &v16, 2);

	bptr = buffer + 12;

	for(i = 0; i < txn->obj_count; i++) {
		v32 = htonl(txn->object[i].obj_len);
		memcpy(bptr, &v32, 4);

		v16 = htons(txn->object[i].obj_id);
		memcpy(bptr + 4, &v16, 2);

		memcpy(bptr + 6, txn->object[i].obj_data, txn->object[i].obj_len);

		bptr += 6 + txn->object[i].obj_len;
	}

	ret = send(conn->sock, buffer, txn->length + 4, 0);
	free(buffer);

	return ret;
#endif 
}
