#ifndef TRANSACTION_H
#define TRANSACTION_H

/** 
 * @file transaction.h
 * @brief Code for constructing and processing transactions
 */
#include "machdep.h"
#include "pdtp.h"

#include "multiplexer.h"
#include "notifier.h"

typedef struct pdtp_transaction_manager *pdtp_transaction_manager_t;
typedef struct pdtp_transaction_listener *pdtp_transaction_listener_t;
typedef struct pdtp_transaction_in *pdtp_transaction_in_t;
typedef struct pdtp_transaction_out *pdtp_transaction_out_t;

typedef enum {
	TXN_COMMAND,
	TXN_REPLY
} pdtp_transaction_type_t;

/** The transaction manager structure stores connection specific information
   that allows multiple threads to utilize a single connection in a thread
   safe manner. */
pdtp_transaction_manager_t pdtp__transaction_manager_create(void);

/** destroy the transaction manager */
void pdtp__transaction_manager_destroy(pdtp_transaction_manager_t txnmgr);

/** An object notifier is used for receiving server originated (and therefore
   asynchronous) transactions.  Object notifiers remain within the
   transaction filtering system until they are explicitly removed with
   pdtp__remove_object_notifier(); */
pdtp_notifier_t pdtp__create_object_notifier(pdtp_connection_t conn, uint16_t id, void *payload, unsigned length);

/** This destroys an object notifier and its associated transaction filter */
void pdtp__destroy_object_notifier(pdtp_connection_t, pdtp_notifier_t notifier);

/** This waits for a transaction matching the filter rules assigned to the
   given notifier. */
pdtp_transaction_in_t pdtp__transaction_wait(pdtp_connection_t, pdtp_notifier_t, pdtp_read_mode_t);

/** This function returns a handle to a multiplexable transaction listener
   object.  It may be added to a multiplexer object using the functions below.
   This listener will be automatically garbage collected when used in 
   conjunction with a serial notifier.  For object notifiers, it must
   explicity be removed from the multiplexer and its resources freed
   with a pdtp__transaction_listener_destroy() call. */
pdtp_transaction_listener_t pdtp__transaction_listener_create(pdtp_connection_t, pdtp_notifier_t, pdtp_read_mode_t);

/** This function destroys a transaction listener object */
void pdtp__transaction_listener_destroy(pdtp_transaction_listener_t);

/** This function adds a transaction listener object to a multiplexer */
void pdtp__multiplexer_add_transaction_listener(pdtp_multiplexer_t, pdtp_transaction_listener_t, int (*function)(pdtp_multiplexer_t, pdtp_transaction_listener_t, pdtp_transaction_in_t, void *arg), void *arg);

/** This function removes a transaction listener object from a multiplexer.
   The arg pointer passed to  the pdtp__multiplexer_add_transaction_listener()
   function originally is returned */
void *pdtp__multiplexer_remove_transaction_listener(pdtp_multiplexer_t, pdtp_transaction_listener_t);

void pdtp__transaction_in_destroy(pdtp_transaction_in_t);
pdtp_transaction_type_t pdtp__transaction_type(pdtp_transaction_in_t);
uint16_t pdtp__transaction_id(pdtp_transaction_in_t);
uint32_t pdtp__transaction_serial(pdtp_transaction_in_t);
unsigned pdtp__transaction_object_count(pdtp_transaction_in_t);
int pdtp__transaction_object_index(pdtp_transaction_in_t, uint16_t type);
int pdtp__transaction_object_id(pdtp_transaction_in_t, unsigned objno, uint16_t *id);
int pdtp__transaction_object_length(pdtp_transaction_in_t, unsigned objno, uint32_t *length);
int pdtp__transaction_object_data(pdtp_transaction_in_t, unsigned objno, void **data);
int pdtp__transaction_object_uint16(pdtp_transaction_in_t, unsigned objno, uint16_t *object);
int pdtp__transaction_object_uint32(pdtp_transaction_in_t, unsigned objno, uint32_t *object);
int pdtp__transaction_object_uint64(pdtp_transaction_in_t, unsigned objno, uint64_t *object);
int pdtp__transaction_object_string(pdtp_transaction_in_t, unsigned objno, char **string);
int pdtp__transaction_object_unpack(pdtp_transaction_in_t, unsigned objno, const char *format, ...);
	
pdtp_transaction_out_t pdtp__transaction_create(uint16_t tid, uint16_t obj_count);
void pdtp__transaction_out_destroy(pdtp_transaction_out_t);
pdtp_notifier_t pdtp__transaction_set_serial(pdtp_transaction_out_t txn, pdtp_connection_t conn);
void pdtp__transaction_add_object(pdtp_transaction_out_t, uint16_t id, const void *data, uint32_t length);
void pdtp__transaction_add_int16(pdtp_transaction_out_t, uint16_t id, uint16_t value);
void pdtp__transaction_add_int32(pdtp_transaction_out_t, uint16_t id, uint32_t value);
void pdtp__transaction_add_string(pdtp_transaction_out_t, uint16_t id, const char *string);
int pdtp__transaction_object_pack(pdtp_transaction_out_t, uint16_t id, const char *format, ...);
int pdtp__transaction_write(pdtp_connection_t, pdtp_transaction_out_t);

#endif /* TRANSACTION_H */
